/*
 * Object.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using System.Security.Policy;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("95dc42f9-d1f9-467b-acdc-3312d4bcdfee")]
    [CommandFlags(CommandFlags.Unsafe)]
    [ObjectGroup("managedEnvironment")]
    internal sealed class Object : Core
    {
        public Object(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IEnsemble Members
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] {
            "alias", "assemblies", "callbackflags", "certificate", "cleanup",
            "create", "declare", "dispose", "exists", "flags",
            "foreach", "get", "hash", "import", "interfaces",
            "invoke", "invokeall", "list", "load", "members",
            "namespaces", "refcount", "removecallback", "resolve", "strongname",
            "type", "types", "unalias", "undeclare", "unimport",
            "untype", "verifyall"
        });

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            switch (subCommand)
                            {
                                case "alias":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) == arguments.Count))
                                                {
                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    IObject @object = null;
                                                    Type type = null;

                                                    if (interpreter.GetObject(arguments[argumentIndex], true, true,
                                                            ref @object) == ReturnCode.Ok)
                                                    {
                                                        if (@object.Value != null)
                                                        {
                                                            type = @object.Value.GetType();
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "invalid value for object \"{0}\"",
                                                                arguments[argumentIndex]);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        ResultList errors = null;

                                                        code = Value.GetType(interpreter, arguments[argumentIndex],
                                                            interpreter.GetAppDomain(), strictType, verbose, noCase,
                                                            interpreter.CultureInfo, ref type, ref errors);

                                                        if (code != ReturnCode.Ok)
                                                        {
                                                            errors.Insert(0, String.Format(
                                                                "object or type \"{0}\" not found",
                                                                arguments[argumentIndex]));

                                                            result = errors;
                                                        }
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        IAlias alias = null;

                                                        code = interpreter.AddObjectAlias(
                                                            arguments[argumentIndex],
                                                            RuntimeOps.GetInvokeOptions(aliasAll),
                                                            aliasAll, ref alias, ref result);

                                                        if (code == ReturnCode.Ok)
                                                            result = alias.ToString();
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object alias ?options? object\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object alias ?options? object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "assemblies":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            AppDomain appDomain = interpreter.GetAppDomain();

                                            if (appDomain != null)
                                            {
                                                StringList list = new StringList(
                                                    appDomain.GetAssemblies());

                                                string pattern = null;

                                                if (arguments.Count == 3)
                                                    pattern = arguments[2];

                                                result = list.ToString(pattern, false);
                                            }
                                            else
                                            {
                                                result = "invalid application domain";
                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object assemblies ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "callbackflags":
                                    {
                                        if ((arguments.Count == 3) || (arguments.Count == 4))
                                        {
                                            ICallback callback = null;

                                            code = interpreter.GetCallback(arguments[2], false, true, ref callback, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                //
                                                // HACK: This following causes a compiler error with the C#
                                                //       v2.0 compiler (i.e. the one included with the .NET
                                                //       Framework 2.0 SP2) because the CallbackFlags enum
                                                //       has a value named "ToString":
                                                //
                                                //       callback.Flags.ToString()
                                                //
                                                //       Therefore, use a temporary object as a workaround.
                                                //
                                                object flags = callback.Flags;

                                                if (arguments.Count == 4)
                                                {
                                                    object enumValue = EnumOps.TryParseFlagsEnum(
                                                        interpreter, typeof(CallbackFlags),
                                                        flags.ToString(),
                                                        arguments[3], interpreter.CultureInfo,
                                                        true, ref result);

                                                    if (enumValue is CallbackFlags)
                                                        callback.Flags = (CallbackFlags)enumValue;
                                                    else
                                                        code = ReturnCode.Error;
                                                }

                                                if (code == ReturnCode.Ok)
                                                {
                                                    //
                                                    // HACK: See above comment for details on this.
                                                    //
                                                    flags = callback.Flags;
                                                    result = flags.ToString();
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object callbackflags name ?flags?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "certificate":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-chain", null),
                                                new Option(typeof(X509VerificationFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-verificationflags",
                                                        new Variant(CertificateOps.DefaultVerificationFlags)),
                                                new Option(typeof(X509RevocationMode), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-revocationmode", null),
                                                new Option(typeof(X509RevocationFlag), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-revocationflag", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) &&
                                                    ((argumentIndex + 1) == arguments.Count))
                                                {
                                                    bool chain = false;

                                                    if (options.IsPresent("-chain"))
                                                        chain = true;

                                                    Variant value = null;
                                                    X509VerificationFlags verificationFlags =
                                                        CertificateOps.DefaultVerificationFlags;

                                                    if (options.IsPresent("-verificationflags", ref value))
                                                        verificationFlags = (X509VerificationFlags)value.Value;

                                                    X509RevocationMode revocationMode =
                                                        X509RevocationMode.Online;

                                                    if (options.IsPresent("-revocationmode", ref value))
                                                        revocationMode = (X509RevocationMode)value.Value;

                                                    X509RevocationFlag revocationFlag =
                                                        X509RevocationFlag.ExcludeRoot;

                                                    if (options.IsPresent("-revocationflag", ref value))
                                                        revocationFlag = (X509RevocationFlag)value.Value;

                                                    IObject @object = null;

                                                    code = interpreter.GetObject(
                                                        arguments[argumentIndex], true, true,
                                                        ref @object, ref result);

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        Assembly assembly = (@object != null) ?
                                                            @object.Value as Assembly : null;

                                                        if (chain)
                                                        {
                                                            if (assembly != null)
                                                            {
                                                                X509Certificate2 certificate = null;

                                                                code = AssemblyOps.GetCertificate2(
                                                                    assembly, true, ref certificate, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                    code = CertificateOps.VerifyChain(
                                                                        assembly, certificate, verificationFlags,
                                                                        revocationMode, revocationFlag, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                    result = FormatOps.Certificate(
                                                                        assembly, certificate, true, true, false);
                                                            }
                                                            else
                                                            {
                                                                result = "invalid assembly";
                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            X509Certificate certificate = null;

                                                            code = AssemblyOps.GetCertificate(
                                                                assembly, ref certificate, ref result);

                                                            if (code == ReturnCode.Ok)
                                                                result = FormatOps.Certificate(
                                                                    assembly, certificate, true, true, false);
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object certificate ?options? assembly\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object certificate ?options? assembly\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "cleanup":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-synchronous", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocomplain", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex == Index.Invalid)
                                                {
                                                    bool synchronous = false;

                                                    if (options.IsPresent("-synchronous"))
                                                        synchronous = true;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool stopOnError = true;

                                                    if (options.IsPresent("-nocomplain"))
                                                        stopOnError = false;

                                                    code = interpreter.RemoveObjects(
                                                        null, stopOnError, synchronous, ref dispose, ref result);
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object cleanup ?options?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object cleanup ?options?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "create":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = RuntimeOps.GetCreateOptions();
                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex != Index.Invalid)
                                                {
                                                    bool trace = false;

                                                    if (options.IsPresent("-trace"))
                                                        trace = true;

                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    bool @default = false;

                                                    if (options.IsPresent("-default"))
                                                        @default = true;

                                                    //
                                                    // NOTE: Prevent any magical type searches (i.e. use their specified
                                                    //       type string verbatim)?
                                                    //
                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    bool strictMember = false;

                                                    if (options.IsPresent("-strictmember"))
                                                        strictMember = true;

                                                    bool strictArgs = false;

                                                    if (options.IsPresent("-strictargs"))
                                                        strictArgs = true;

                                                    //
                                                    // NOTE: Return all Type exception information (this can be very
                                                    //       costly for performance).
                                                    //
                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    //
                                                    // NOTE: Do we expect to create an object (e.g. System.Int32.Parse)?
                                                    //
                                                    bool create = true;

                                                    if (options.IsPresent("-nocreate"))
                                                        create = false;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool invoke = true;

                                                    if (options.IsPresent("-noinvoke"))
                                                        invoke = false;

                                                    bool noArgs = false;

                                                    if (options.IsPresent("-noargs"))
                                                        noArgs = true;

                                                    //
                                                    // NOTE: Do we want to create a command alias for the newly created
                                                    //       object?
                                                    //
                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    //
                                                    // NOTE: Forcibly convert the object to a string and discard it?
                                                    //
                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    Variant value = null;
                                                    ByRefArgumentFlags argumentFlags = MarshalOps.DefaultByRefArgumentFlags;

                                                    if (options.IsPresent("-argumentflags", ref value))
                                                        argumentFlags = (ByRefArgumentFlags)value.Value;

                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    TypeList parameterTypes = null;

                                                    if (options.IsPresent("-parametertypes", ref value))
                                                        parameterTypes = (TypeList)value.Value;

                                                    BindingFlags bindingFlags = MarshalOps.DefaultBindingFlags;

                                                    if (options.IsPresent("-flags", ref value))
                                                        bindingFlags = (BindingFlags)value.Value;

                                                    int limit = invoke ? 1 : 0;

                                                    if (options.IsPresent("-limit", ref value))
                                                        limit = (int)value.Value;

                                                    int index = Index.Invalid;

                                                    if (options.IsPresent("-index", ref value))
                                                        index = (int)value.Value;

                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    //
                                                    // NOTE: We are going to modify the interpreter state, make
                                                    //       sure it is not set to read-only.
                                                    //
                                                    if (!interpreter.ReadOnly)
                                                    {
                                                        Type objectType = null;
                                                        ResultList errors = null;

                                                        code = Value.GetType(interpreter, arguments[argumentIndex],
                                                            interpreter.GetAppDomain(), strictType, verbose, noCase,
                                                            interpreter.CultureInfo, ref objectType, ref errors);

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            object[] args = null;
                                                            int argumentCount = 0;

                                                            if ((argumentIndex + 1) < arguments.Count)
                                                            {
                                                                //
                                                                // NOTE: How many arguments were supplied?
                                                                //
                                                                argumentCount = (arguments.Count - (argumentIndex + 1));

                                                                //
                                                                // NOTE: Create and populate the array of arguments for the
                                                                //       invocation.
                                                                //
                                                                args = new object[argumentCount];

                                                                for (int index2 = (argumentIndex + 1); index2 < arguments.Count; index2++)
                                                                    /* need String, not Argument */
                                                                    args[index2 - (argumentIndex + 1)] = arguments[index2].ToString();
                                                            }
                                                            else if (invoke || !noArgs)
                                                            {
                                                                //
                                                                // FIXME: When no arguments are specified, we actually need an array
                                                                //        of zero arguments for the parameter to argument matching
                                                                //        code to work correctly.
                                                                //
                                                                args = new object[0];
                                                            }

                                                            ConstructorInfo[] constructorInfo = objectType.GetConstructors(bindingFlags);

                                                            if ((constructorInfo != null) && (constructorInfo.Length > 0))
                                                            {
                                                                IntList methodIndexList = null;
                                                                ObjectArrayList argsList = null;
                                                                ArgumentInfoList argumentInfoList = null;

                                                                //
                                                                // NOTE: Attempt to convert the argument strings to something
                                                                //       potentially more meaningful and find the corresponding
                                                                //       constructor.
                                                                //
                                                                errors = null;

                                                                code = MarshalOps.FindMethodsAndFixupArguments(
                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                    objectType, arguments[argumentIndex], null, MemberTypes.Constructor,
                                                                    bindingFlags, constructorInfo, parameterTypes, args, limit, verbose,
                                                                    @default, ref methodIndexList, ref argsList, ref argumentInfoList,
                                                                    ref errors);

                                                                if (code == ReturnCode.Ok)
                                                                {
                                                                    if ((methodIndexList != null) && (methodIndexList.Count > 0) &&
                                                                        (argsList != null) && (argsList.Count > 0))
                                                                    {
                                                                        if ((index == Index.Invalid) ||
                                                                            ((index >= 0) && (index < methodIndexList.Count) &&
                                                                            (index < argsList.Count)))
                                                                        {
                                                                            if (invoke)
                                                                            {
                                                                                if (!strictMember || (methodIndexList.Count == 1))
                                                                                {
                                                                                    //
                                                                                    // FIXME: Select the first method that matches OR the one
                                                                                    //        specified by the caller.  More sophisticated
                                                                                    //        logic may need to be added here later.
                                                                                    //
                                                                                    int methodIndex = (index != Index.Invalid) ?
                                                                                        methodIndexList[index] : methodIndexList[0];

                                                                                    if (methodIndex != Index.Invalid)
                                                                                    {
                                                                                        try
                                                                                        {
                                                                                            //
                                                                                            // NOTE: Get the arguments we are going to use to perform
                                                                                            //       the actual method call.
                                                                                            //
                                                                                            args = (index != Index.Invalid) ? argsList[index] : argsList[0];

                                                                                            if (trace)
                                                                                                RuntimeOps.DebugTrace(String.Format(
                                                                                                    "{0}: Execute: methodIndex = {1}, constructorInfo = {2}, " +
                                                                                                    "args = {3}{4}", GlobalState.GetCurrentThreadId(),
                                                                                                    methodIndex, FormatOps.WrapOrNull(constructorInfo[methodIndex]),
                                                                                                    FormatOps.WrapOrNull(new StringList(args)),
                                                                                                    Environment.NewLine), typeof(Object).Name);

                                                                                            object @object = constructorInfo[methodIndex].Invoke(
                                                                                                bindingFlags, interpreter.Binder as Binder, args, interpreter.CultureInfo);

                                                                                            if (@object != null)
                                                                                            {
                                                                                                if (argumentInfoList != null)
                                                                                                    code = MarshalOps.FixupByRefArguments(
                                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                        argumentInfoList, objectFlags,
                                                                                                        RuntimeOps.GetInvokeOptions(aliasAll), interpName, args,
                                                                                                        argumentFlags, strictArgs, create, dispose, alias,
                                                                                                        aliasAll, toString, ref result);

                                                                                                if (code == ReturnCode.Ok)
                                                                                                    code = MarshalOps.FixupReturnValue(
                                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                        type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                        objectName, interpName, @object, create, dispose, alias,
                                                                                                        aliasAll, toString, ref result);
                                                                                            }
                                                                                            else
                                                                                            {
                                                                                                result = String.Format(
                                                                                                    "could not create an instance of type \"{0}\"",
                                                                                                    objectType.FullName);

                                                                                                code = ReturnCode.Error;
                                                                                            }
                                                                                        }
                                                                                        catch (Exception e)
                                                                                        {
                                                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                                                            result = e;
                                                                                            code = ReturnCode.Error;
                                                                                        }
                                                                                    }
                                                                                    else
                                                                                    {
                                                                                        result = String.Format(
                                                                                            "type \"{0}\" constructor not found",
                                                                                            objectType.FullName);

                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                }
                                                                                else
                                                                                {
                                                                                    result = String.Format(
                                                                                        "matched {0} constructors overloads on type \"{1}\", need exactly 1",
                                                                                         methodIndexList.Count, objectType.FullName);

                                                                                    code = ReturnCode.Error;
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                ConstructorInfoList constructorInfoList = new ConstructorInfoList();

                                                                                if (index != Index.Invalid)
                                                                                    constructorInfoList.Add(constructorInfo[methodIndexList[index]]);
                                                                                else
                                                                                    foreach (int methodIndex in methodIndexList)
                                                                                        constructorInfoList.Add(constructorInfo[methodIndex]);

                                                                                code = MarshalOps.FixupReturnValue(
                                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                    type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                    objectName, interpName, constructorInfoList, create,
                                                                                    dispose, alias, aliasAll, toString, ref result);
                                                                            }
                                                                        }
                                                                        else
                                                                        {
                                                                            result = String.Format(
                                                                                "type \"{0}\" constructor not found, " +
                                                                                "invalid method index {1}, must be {2}",
                                                                                objectType.FullName, index,
                                                                                FormatOps.BetweenOrExact(0, methodIndexList.Count - 1));

                                                                            code = ReturnCode.Error;
                                                                        }
                                                                    }
                                                                    else
                                                                    {
                                                                        result = String.Format(
                                                                            "type \"{0}\" constructor not found",
                                                                            objectType.FullName);

                                                                        code = ReturnCode.Error;
                                                                    }
                                                                }
                                                                else
                                                                {
                                                                    result = errors;
                                                                }
                                                            }
                                                            //
                                                            // NOTE: These (primitive and value types) may not have any
                                                            //       constructors (e.g. System.Int32).
                                                            //
                                                            else if (invoke && (objectType.IsPrimitive || objectType.IsValueType))
                                                            {
                                                                if (index == Index.Invalid)
                                                                {
                                                                    try
                                                                    {
                                                                        object @object = Activator.CreateInstance(
                                                                            objectType, bindingFlags, interpreter.Binder as Binder, args,
                                                                            interpreter.CultureInfo);

                                                                        if (@object != null)
                                                                        {
                                                                            code = MarshalOps.FixupReturnValue(
                                                                                interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                objectName, interpName, @object, create, dispose, alias,
                                                                                aliasAll, toString, ref result);
                                                                        }
                                                                        else
                                                                        {
                                                                            result = String.Format(
                                                                                "could not create an instance of primitive type \"{0}\"",
                                                                                objectType.FullName);

                                                                            code = ReturnCode.Error;
                                                                        }
                                                                    }
                                                                    catch (Exception e)
                                                                    {
                                                                        Engine.SetExceptionErrorCode(interpreter, e);

                                                                        result = e;
                                                                        code = ReturnCode.Error;
                                                                    }
                                                                }
                                                                else
                                                                {
                                                                    result = String.Format(
                                                                        "type \"{0}\" constructor not found, " +
                                                                        "cannot specify method index {1} for primitive type",
                                                                        objectType.FullName, index);

                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "type \"{0}\" has no constructors matching \"{1}\"",
                                                                    objectType.FullName, bindingFlags);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            errors.Insert(0, String.Format(
                                                                "type \"{0}\" not found",
                                                                arguments[argumentIndex]));

                                                            result = errors;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = "interpreter is read-only";
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object create ?options? typeName ?arg ...?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object create ?options? typeName ?arg ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "declare":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declarenonpublic", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-declaremode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-declarepattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declarenocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: Strict handling for type names?
                                                //
                                                bool strictType = false;

                                                if (options.IsPresent("-stricttype"))
                                                    strictType = true;

                                                //
                                                // NOTE: Return all Type exception information (this can be very
                                                //       costly for performance).
                                                //
                                                bool verbose = false;

                                                if (options.IsPresent("-verbose"))
                                                    verbose = true;

                                                //
                                                // NOTE: Do we want to declare non-public interfaces as well?
                                                //
                                                bool nonPublic = false;

                                                if (options.IsPresent("-declarenonpublic"))
                                                    nonPublic = true;

                                                //
                                                // NOTE: What match mode are we going to use to restrict the declared
                                                //       interfaces, if any?
                                                //
                                                Variant value = null;
                                                MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                if (options.IsPresent("-declaremode", ref value))
                                                    mode = (MatchMode)value.Value;

                                                //
                                                // NOTE: What pattern are we going to use to restrict the declared
                                                //       interfaces, if any?
                                                //
                                                string pattern = null;

                                                if (options.IsPresent("-declarepattern", ref value))
                                                    pattern = value.ToString();

                                                //
                                                // NOTE: For the purposes of matching interface names, do we care
                                                //       about case?
                                                //
                                                bool noCase = false;

                                                if (options.IsPresent("-declarenocase"))
                                                    noCase = true;

                                                //
                                                // NOTE: Check for a pattern without a mode (change to default,
                                                //       which is Glob).
                                                //
                                                if ((pattern != null) && (mode == MatchMode.None))
                                                    mode = StringOps.DefaultObjectMatchMode;

                                                //
                                                // NOTE: Figure out which interfaces they want to declare (add).
                                                //
                                                TypeList types = new TypeList();

                                                //
                                                // NOTE: Any specific interfaces they want to declare (add)?
                                                //
                                                if (argumentIndex != Index.Invalid)
                                                {
                                                    StringList list = new StringList(ArgumentList.GetRange(
                                                        arguments, argumentIndex, Index.Invalid));

                                                    ResultList errors = null;

                                                    code = Value.GetTypeList(interpreter, list.ToString(),
                                                        interpreter.GetAppDomain(), strictType, verbose, noCase,
                                                        interpreter.CultureInfo, ref types, ref errors);

                                                    if (code != ReturnCode.Ok)
                                                        result = errors;
                                                }

                                                if (code == ReturnCode.Ok)
                                                {
                                                    //
                                                    // NOTE: Ok, add the specified interfaces now.
                                                    //
                                                    code = interpreter.AddObjectInterfaces(types, nonPublic, mode,
                                                        pattern, noCase, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        result = String.Empty;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object declare ?options? ?name name ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "dispose":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = RuntimeOps.GetDisposeOptions();
                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex != Index.Invalid)
                                                {
                                                    bool synchronous = false;

                                                    if (options.IsPresent("-synchronous"))
                                                        synchronous = true;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool stopOnError = true;

                                                    if (options.IsPresent("-nocomplain"))
                                                        stopOnError = false;

                                                    int removed = 0;
                                                    int disposed = 0;

                                                    for (; argumentIndex < arguments.Count; argumentIndex++)
                                                    {
                                                        bool localDispose = dispose;
                                                        Result localResult = null;

                                                        code = interpreter.RemoveObject(
                                                            arguments[argumentIndex], null, synchronous,
                                                            ref localDispose, ref localResult);

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            removed++;

                                                            if (localDispose)
                                                                disposed++;
                                                        }
                                                        else if (stopOnError)
                                                        {
                                                            result = localResult;
                                                            break;
                                                        }
                                                        else
                                                        {
                                                            //
                                                            // NOTE: The "-nocomplain" option is enabled;
                                                            //       therefore, reset the return code to
                                                            //       success just in case we exit the loop.
                                                            //
                                                            code = ReturnCode.Ok;
                                                        }
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        if (removed > 1)
                                                            result = StringList.MakeList(
                                                                "disposed", disposed, "removed", removed);
                                                        else if (disposed > 0)
                                                            result = "disposed";
                                                        else if (removed > 0)
                                                            result = "removed";
                                                        else
                                                            result = String.Empty;
                                                    }
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object dispose ?options? ?object ...?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object dispose ?options? ?object ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "exists":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            result = (interpreter.DoesObjectExist(arguments[2]) == ReturnCode.Ok);
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object exists object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "flags":
                                    {
                                        if ((arguments.Count == 3) || (arguments.Count == 4))
                                        {
                                            IObject @object = null;

                                            code = interpreter.GetObject(arguments[2], true, true, ref @object, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (arguments.Count == 4)
                                                {
                                                    object enumValue = EnumOps.TryParseFlagsEnum(
                                                        interpreter, typeof(ObjectFlags),
                                                        @object.Flags.ToString(),
                                                        arguments[3], interpreter.CultureInfo,
                                                        true, ref result);

                                                    if (enumValue is ObjectFlags)
                                                    {
                                                        //
                                                        // NOTE: We are going to modify the interpreter state, make
                                                        //       sure it is not set to read-only.
                                                        //
                                                        if (!interpreter.ReadOnly)
                                                        {
                                                            @object.Flags = (ObjectFlags)enumValue;
                                                        }
                                                        else
                                                        {
                                                            result = "interpreter is read-only";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        code = ReturnCode.Error;
                                                    }
                                                }

                                                if (code == ReturnCode.Ok)
                                                    result = @object.Flags;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object flags object ?flags?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "foreach":
                                    {
                                        if (arguments.Count >= 5)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocreate", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) &&
                                                    ((argumentIndex + 3) == arguments.Count))
                                                {
                                                    bool create = true;

                                                    if (options.IsPresent("-nocreate"))
                                                        create = false;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    Variant value = null;
                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    IObject @object = null;

                                                    code = interpreter.GetObject(
                                                        arguments[argumentIndex + 1], true,
                                                        true, ref @object, ref result);

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        IEnumerable enumerable = (@object != null) ?
                                                            @object.Value as IEnumerable : null;

                                                        if (enumerable != null)
                                                        {
                                                            IEnumerator enumerator = null;

                                                            try
                                                            {
                                                                enumerator = enumerable.GetEnumerator(); /* throw */
                                                            }
                                                            catch
                                                            {
                                                                // do nothing.
                                                            }

                                                            if (enumerator != null)
                                                            {
                                                                string varName = arguments[argumentIndex];
                                                                string body = arguments[argumentIndex + 2];
                                                                IScriptLocation location = arguments[argumentIndex + 2];

                                                                //
                                                                // NOTE: Move to the next item.  If this fails, there are
                                                                //       no more items and we cannot move any farther.
                                                                //
                                                                while ((code == ReturnCode.Ok) && enumerator.MoveNext())
                                                                {
                                                                    //
                                                                    // NOTE: Get the value of the current item.
                                                                    //
                                                                    object newValue = enumerator.Current;

                                                                    //
                                                                    // NOTE: Now create an object handle for it.
                                                                    //
                                                                    code = MarshalOps.FixupReturnValue(
                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                        type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                        objectName, interpName, newValue, create, dispose, alias,
                                                                        aliasAll, toString, ref result);

                                                                    //
                                                                    // NOTE: If we fail at adding the object handle, bail
                                                                    //       out now.
                                                                    //
                                                                    if (code != ReturnCode.Ok)
                                                                    {
                                                                        Engine.AddErrorInformation(interpreter, result,
                                                                            String.Format("{0}    (adding object foreach object handle \"{1}\"",
                                                                                Environment.NewLine, FormatOps.Ellipsis(varName)));
                                                                        break;
                                                                    }

                                                                    //
                                                                    // NOTE: Save the object name just in case we need to
                                                                    //       dispose it (below).
                                                                    //
                                                                    string newObjectName = result;

                                                                    //
                                                                    // NOTE: Set the value of the variable to the created
                                                                    //       object handle.
                                                                    //
                                                                    code = interpreter.SetVariableValue(
                                                                        VariableFlags.None, varName, newObjectName, null, ref result);

                                                                    //
                                                                    // NOTE: If we fail at setting the loop variable, bail
                                                                    //       out now.
                                                                    //
                                                                    if (code != ReturnCode.Ok)
                                                                    {
                                                                        ReturnCode removeCode = ReturnCode.Ok;
                                                                        Result removeResult = null;

                                                                        removeCode = interpreter.RemoveObject(
                                                                            newObjectName, null, true, ref removeResult);

                                                                        if (removeCode != ReturnCode.Ok)
                                                                            DebugOps.Complain(interpreter, removeCode, removeResult);

                                                                        Engine.AddErrorInformation(interpreter, result,
                                                                            String.Format("{0}    (setting object foreach loop variable \"{1}\"",
                                                                                Environment.NewLine, FormatOps.Ellipsis(varName)));
                                                                        break;
                                                                    }

                                                                    //
                                                                    // NOTE: Evaluate the loop body.  If this fails, we bail
                                                                    //       out, leaving the loop variable and associated
                                                                    //       object handle untouched.
                                                                    //
                                                                    code = interpreter.EvaluateScript(body, location, ref result);

                                                                    if (code != ReturnCode.Ok)
                                                                    {
                                                                        if (code == ReturnCode.Continue)
                                                                        {
                                                                            code = ReturnCode.Ok;
                                                                        }
                                                                        else if (code == ReturnCode.Break)
                                                                        {
                                                                            code = ReturnCode.Ok;
                                                                            break;
                                                                        }
                                                                        else if (code == ReturnCode.Error)
                                                                        {
                                                                            Engine.AddErrorInformation(interpreter, result,
                                                                                String.Format("{0}    (\"object foreach\" body line {1})",
                                                                                    Environment.NewLine, interpreter.ErrorLine));
                                                                            break;
                                                                        }
                                                                        else
                                                                        {
                                                                            break;
                                                                        }
                                                                    }
                                                                }

                                                                //
                                                                // NOTE: Upon success, clear result.
                                                                //
                                                                if (code == ReturnCode.Ok)
                                                                    Engine.ResetResult(interpreter, ref result);
                                                            }
                                                            else
                                                            {
                                                                result = "invalid object enumerator";
                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = "object does not support IEnumerable";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object foreach varName object body\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object foreach varName object body\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "get":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocreate", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) &&
                                                    ((argumentIndex + 2) <= arguments.Count) &&
                                                    ((argumentIndex + 3) >= arguments.Count))
                                                {
                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    //
                                                    // NOTE: Prevent any magical type searches (i.e. use their specified
                                                    //       type string verbatim)?
                                                    //
                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    //
                                                    // NOTE: Return all Type exception information (this can be very
                                                    //       costly for performance).
                                                    //
                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    //
                                                    // NOTE: Do we expect to create an object (e.g. System.Int32.Parse)?
                                                    //
                                                    bool create = true;

                                                    if (options.IsPresent("-nocreate"))
                                                        create = false;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    //
                                                    // NOTE: Do we want to create a command alias for the newly created
                                                    //       object?
                                                    //
                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    //
                                                    // NOTE: Forcibly convert the object to a string and discard it?
                                                    //
                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    //
                                                    // NOTE: Which Tcl interpreter do we want a command alias created
                                                    //       in, if any?
                                                    //
                                                    Variant value = null;
                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    Type objectType = null;
                                                    ResultList errors = null;

                                                    code = Value.GetType(interpreter, arguments[argumentIndex],
                                                        interpreter.GetAppDomain(), strictType, verbose, noCase,
                                                        interpreter.CultureInfo, ref objectType, ref errors);

                                                    if (code != ReturnCode.Ok)
                                                    {
                                                        errors.Insert(0, String.Format(
                                                            "object or type \"{0}\" not found",
                                                            arguments[argumentIndex]));

                                                        result = errors;
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        IObject @object = null;

                                                        if ((argumentIndex + 3) == arguments.Count)
                                                            code = interpreter.GetObject(arguments[argumentIndex + 2],
                                                                true, true, ref @object, ref result);

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            try
                                                            {
                                                                object returnValue = Activator.GetObject(
                                                                    objectType, arguments[argumentIndex + 1],
                                                                    (@object != null) ? @object.Value : null);

                                                                code = MarshalOps.FixupReturnValue(
                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                    type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                    objectName, interpName, returnValue, create, dispose,
                                                                    alias, aliasAll, toString, ref result);
                                                            }
                                                            catch (Exception e)
                                                            {
                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                result = e;
                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object get ?options? type url ?state?\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object get ?options? type url ?state?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "hash":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            IObject @object = null;

                                            code = interpreter.GetObject(arguments[2], true, true, ref @object, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                Assembly assembly = (@object != null) ? @object.Value as Assembly : null;
                                                Hash hash = null;

                                                code = AssemblyOps.GetHash(assembly, ref hash, ref result);

                                                if (code == ReturnCode.Ok)
                                                    result = FormatOps.Hash(hash);
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object hash assembly\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "import":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-system", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-assembly", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-container", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-importmode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-importpattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-importnocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: What is the logical container for the manually added
                                                //       namespaces?
                                                //
                                                Variant value = null;
                                                string container = null;

                                                if (options.IsPresent("-container", ref value))
                                                    container = value.ToString();

                                                //
                                                // NOTE: What match mode are we going to use to restrict the imported
                                                //       namespaces, if any?
                                                //
                                                MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                if (options.IsPresent("-importmode", ref value))
                                                    mode = (MatchMode)value.Value;

                                                //
                                                // NOTE: What pattern are we going to use to restrict the imported
                                                //       namespaces, if any?
                                                //
                                                string pattern = null;

                                                if (options.IsPresent("-importpattern", ref value))
                                                    pattern = value.ToString();

                                                //
                                                // NOTE: For the purposes of matching namespace names, do we care
                                                //       about case?
                                                //
                                                bool noCase = false;

                                                if (options.IsPresent("-importnocase"))
                                                    noCase = true;

                                                bool assembly = false;

                                                if (options.IsPresent("-assembly"))
                                                    assembly = true;

                                                bool system = false;

                                                if (options.IsPresent("-system"))
                                                    system = true;

                                                //
                                                // NOTE: Check for a pattern without a mode (change to default,
                                                //       which is Glob).
                                                //
                                                if ((pattern != null) && (mode == MatchMode.None))
                                                    mode = StringOps.DefaultObjectMatchMode;

                                                //
                                                // NOTE: Figure out which namespaces they want to import (add).
                                                //
                                                StringDictionary dictionary = new StringDictionary();

                                                if ((code == ReturnCode.Ok) && assembly)
                                                    code = RuntimeOps.AddAssemblyObjectNamespaces(ref dictionary, ref result);

                                                if ((code == ReturnCode.Ok) && system)
                                                    code = RuntimeOps.AddSystemObjectNamespaces(ref dictionary, ref result);

                                                if (code == ReturnCode.Ok)
                                                {
                                                    //
                                                    // NOTE: Any specific namespaces they want to import (add)?
                                                    //
                                                    if (argumentIndex != Index.Invalid)
                                                        dictionary.AddKeys(ArgumentList.GetRange(
                                                            arguments, argumentIndex, Index.Invalid), container);

                                                    //
                                                    // NOTE: Ok, add the specified namespaces now.
                                                    //
                                                    code = interpreter.AddObjectNamespaces(
                                                        dictionary, mode, pattern, noCase, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        result = String.Empty;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object import ?options? ?name name ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "interfaces":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            lock (interpreter.SyncRoot)
                                            {
                                                string pattern = null;

                                                if (arguments.Count == 3)
                                                    pattern = arguments[2];

                                                TypePairDictionary<string, long> dictionary = interpreter.ObjectInterfaces;

                                                if (dictionary != null)
                                                    result = dictionary.KeysAndValuesToString(pattern, false);
                                                else
                                                    result = String.Empty;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object interfaces ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "invoke":
                                    {
                                        //
                                        // FIXME: There are a variety of things we could do to make invoking
                                        //        non-static objects via a simplified syntax possible (e.g.
                                        //        $x ToString):
                                        //
                                        //        1. The engine could be modified to look for objects in the
                                        //           interpreter (via GetObject on the first word of the
                                        //           command) and automatically consider them eligible to
                                        //           be executed as "commands".  They would then vector to
                                        //           the [object invoke] code (either by the engine supporting
                                        //           dynamic command re-writing, prefixing "object invoke " in
                                        //           this case, or by simply wrapping the actual [object invoke]
                                        //           code in a SubCommand object that can easily be called both
                                        //           by the engine and for actual [object invoke] calls).
                                        //
                                        //        2. A command could be added to the interpreter upon
                                        //           [object create] that calls the above mentioned
                                        //           [object invoke] SubCommand object passing only the supplied
                                        //           object and arguments (no dynamic command re-writing).
                                        //
                                        // NOTE: The above issue has been "fixed" (use the -alias option).
                                        //
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = RuntimeOps.GetInvokeOptions();
                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false,
                                                ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: We require at least the opaque object handle and member name
                                                //       after the options.
                                                //
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) < arguments.Count))
                                                {
                                                    bool invokeAll = false;

                                                    if (options.IsPresent("-invokeall"))
                                                        invokeAll = true;

                                                    if (invokeAll)
                                                    {
                                                        //
                                                        // NOTE: Copy our current argument list verbatim.
                                                        //
                                                        ArgumentList newArguments = new ArgumentList(arguments);

                                                        //
                                                        // NOTE: Replace the sub-command name with "invokeall".
                                                        //
                                                        newArguments[1] = "invokeall";

                                                        //
                                                        // NOTE: Invoke this method recursively with the modified
                                                        //       argument list.
                                                        //
                                                        return Execute(interpreter, clientData, newArguments, ref result);
                                                    }

                                                    bool trace = false;

                                                    if (options.IsPresent("-trace"))
                                                        trace = true;

                                                    //
                                                    // NOTE: Do we expect to create an object (e.g. System.Int32.Parse)?
                                                    //
                                                    bool create = false;

                                                    if (options.IsPresent("-create"))
                                                        create = true;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool invoke = true;

                                                    if (options.IsPresent("-noinvoke"))
                                                        invoke = false;

                                                    bool noArgs = false;

                                                    if (options.IsPresent("-noargs"))
                                                        noArgs = true;

                                                    //
                                                    // NOTE: Prevent any magical type searches (i.e. use their specified
                                                    //       type string verbatim)?
                                                    //
                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    bool strictMember = false;

                                                    if (options.IsPresent("-strictmember"))
                                                        strictMember = true;

                                                    bool strictArgs = false;

                                                    if (options.IsPresent("-strictargs"))
                                                        strictArgs = true;

                                                    //
                                                    // NOTE: Do we want to create a command alias for any newly created
                                                    //       objects?
                                                    //
                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    //
                                                    // NOTE: Forcibly convert the object to a string and discard it?
                                                    //
                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    //
                                                    // NOTE: Return all Type exception information (this can be very
                                                    //       costly for performance).
                                                    //
                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    bool @default = false;

                                                    if (options.IsPresent("-default"))
                                                        @default = true;

                                                    //
                                                    // NOTE: Prohibit splitting the object/type name argument and
                                                    //       performing a search for a qualified or "nested" object
                                                    //       reference?
                                                    //
                                                    bool noNestedObject = false;

                                                    if (options.IsPresent("-nonestedobject"))
                                                        noNestedObject = true;

                                                    bool noNestedMember = false;

                                                    if (options.IsPresent("-nonestedmember"))
                                                        noNestedMember = true;

                                                    Variant value = null;
                                                    ByRefArgumentFlags argumentFlags =
                                                        MarshalOps.DefaultByRefArgumentFlags;

                                                    if (options.IsPresent("-argumentflags", ref value))
                                                        argumentFlags = (ByRefArgumentFlags)value.Value;

                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    Type returnType = null;

                                                    if (options.IsPresent("-returntype", ref value))
                                                        returnType = (Type)value.Value;

                                                    TypeList parameterTypes = null;

                                                    if (options.IsPresent("-parametertypes", ref value))
                                                        parameterTypes = (TypeList)value.Value;

                                                    MemberTypes memberTypes = MarshalOps.DefaultMemberTypes;

                                                    if (options.IsPresent("-membertypes", ref value))
                                                        memberTypes = (MemberTypes)value.Value;

                                                    BindingFlags bindingFlags = MarshalOps.DefaultBindingFlags;

                                                    if (options.IsPresent("-flags", ref value))
                                                        bindingFlags = (BindingFlags)value.Value;

                                                    int limit = invoke ? 1 : 0;

                                                    if (options.IsPresent("-limit", ref value))
                                                        limit = (int)value.Value;

                                                    int index = Index.Invalid;

                                                    if (options.IsPresent("-index", ref value))
                                                        index = (int)value.Value;

                                                    //
                                                    // NOTE: Which Tcl interpreter do we want a command alias
                                                    //       created in, if any?
                                                    //
                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    bool identity = false;

                                                    if (options.IsPresent("-identity"))
                                                        identity = true;

                                                    //
                                                    // NOTE: Now check for and use the -nocase value.  It is also
                                                    //       important to note here that a specifying the flags
                                                    //       does not override this setting.
                                                    //
                                                    if (noCase)
                                                        bindingFlags |= BindingFlags.IgnoreCase;

                                                    //
                                                    // NOTE: For "safe" interpreters, prevent the use of "unsafe"
                                                    //       member types and binding flags (e.g. NonPublic, etc).
                                                    //
                                                    if (interpreter.IsSafe())
                                                    {
                                                        memberTypes &= ~MarshalOps.UnsafeObjectMemberTypes;
                                                        bindingFlags &= ~MarshalOps.UnsafeObjectBindingFlags;
                                                    }

                                                    //
                                                    // NOTE: Now check for an use the -identity value.  It is also
                                                    //       important to note here that a specifying the flags
                                                    //       does not override this setting.
                                                    //
                                                    if (identity)
                                                        //
                                                        // NOTE: These flags are needed because of the precise
                                                        //       signature of the "MarshalOps.Identity" method.
                                                        //
                                                        bindingFlags |= MarshalOps.PublicStaticMethodBindingFlags;

                                                    TypedInstance typedInstance = null;

                                                    code = Value.GetNestedObject(
                                                        interpreter, arguments[argumentIndex], interpreter.GetAppDomain(),
                                                        bindingFlags, type, strictType, verbose, noCase, noNestedObject,
                                                        interpreter.CultureInfo, ref typedInstance, ref result);

                                                    Type objectType = null;
                                                    object @object = null;

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        objectType = typedInstance.Type;

                                                        if (objectType != null)
                                                        {
                                                            @object = typedInstance.Object;
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "invalid object or type \"{0}\"",
                                                                arguments[argumentIndex]);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }

                                                    //
                                                    // NOTE: Did the type lookup above succeed?
                                                    //
                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        object[] args = null;
                                                        int argumentCount = 0;

                                                        if (identity)
                                                        {
                                                            //
                                                            // NOTE: In "self" mode, the object itself is always the one
                                                            //       and only argument.
                                                            //
                                                            args = new object[] { @object };

                                                            //
                                                            // NOTE: There should only be one argument; however, query
                                                            //       the create array anyhow just in case we change it.
                                                            //
                                                            argumentCount = args.Length;
                                                        }
                                                        else if ((argumentIndex + 2) < arguments.Count)
                                                        {
                                                            //
                                                            // NOTE: How many arguments were supplied?
                                                            //
                                                            argumentCount = (arguments.Count - (argumentIndex + 2));

                                                            //
                                                            // NOTE: Create and populate the array of arguments for the
                                                            //       invocation.
                                                            //
                                                            args = new object[argumentCount];

                                                            for (int index2 = (argumentIndex + 2); index2 < arguments.Count; index2++)
                                                                /* need String, not Argument */
                                                                args[index2 - (argumentIndex + 2)] = arguments[index2].ToString();
                                                        }
                                                        else if (invoke || !noArgs)
                                                        {
                                                            //
                                                            // FIXME: When no arguments are specified, we actually need an array
                                                            //        of zero arguments for the parameter to argument matching
                                                            //        code to work correctly.
                                                            //
                                                            args = new object[0];
                                                        }

                                                        //
                                                        // FIXME: This is not quite right.  We are filtering based on member
                                                        //        type and binding flags and then selecting the first match
                                                        //        to proceed with processing on.  It would be nicer if we
                                                        //        could skip this step altogether; however, we have to know
                                                        //        what kind of member we are dealing with before we can do
                                                        //        anything meaningful with it (and preferably, without having
                                                        //        to get this information from the script writer).
                                                        //
                                                        TypedMember typedMemberInfo = null;

                                                        if (identity)
                                                        {
                                                            //
                                                            // NOTE: While the member name must be present in the object
                                                            //       invoke call (primarily due to the argument count
                                                            //       checking and the [normally totally valid] hard-wired
                                                            //       assumptions about argument indexing), it is ignored
                                                            //       when the -identity option is used.
                                                            //
                                                            typedMemberInfo = new TypedMember(
                                                                typeof(MarshalOps), ObjectFlags.None, @object,
                                                                MarshalOps.IdentityMemberName,
                                                                MarshalOps.IdentityMemberInfo);
                                                        }
                                                        else if (!String.IsNullOrEmpty(arguments[argumentIndex + 1]))
                                                            //
                                                            // NOTE: Attempt to resolve the final member information based on
                                                            //       the provided [possibly composite] member name and the
                                                            //       specified member types and binding flags.
                                                            //
                                                            code = Value.GetNestedMember(
                                                                interpreter, arguments[argumentIndex + 1], typedInstance,
                                                                memberTypes, bindingFlags, noNestedMember,
                                                                interpreter.CultureInfo, ref typedMemberInfo, ref result);
                                                        else
                                                            //
                                                            // NOTE: A null or empty string was provided, skip the standard
                                                            //       member name resolution and simply use the default
                                                            //       member(s).
                                                            //
                                                            typedMemberInfo = new TypedMember(
                                                                objectType, ObjectFlags.None, @object,
                                                                arguments[argumentIndex + 1],
                                                                objectType.GetDefaultMembers());

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            //
                                                            // NOTE: The type and object instance could have changed during
                                                            //       the member resolution process (above); therefore,
                                                            //       refresh them from that result now.
                                                            //
                                                            objectType = typedMemberInfo.Type;
                                                            @object = typedMemberInfo.Object;

                                                            //
                                                            // NOTE: Get the object and method names to pass to the method
                                                            //       overload resolution engine.  Also, get the array of
                                                            //       methods to actually perform overload resolution over.
                                                            //
                                                            string newObjectName = typedInstance.ObjectName; /* COSMETIC */
                                                            string newMemberName = typedMemberInfo.MemberName;
                                                            MemberInfo[] memberInfo = typedMemberInfo.MemberInfo;

                                                            //
                                                            // NOTE: Make sure we found some members to perform overload
                                                            //       resolution over.
                                                            //
                                                            if ((memberInfo != null) && (memberInfo.Length > 0))
                                                            {
                                                                switch (memberInfo[0].MemberType)
                                                                {
                                                                    case MemberTypes.Field:
                                                                        {
                                                                            FieldInfo fieldInfo = objectType.GetField(
                                                                                newMemberName, bindingFlags);

                                                                            if (fieldInfo != null)
                                                                            {
                                                                                if (invoke)
                                                                                {
                                                                                    if (argumentCount == 0)
                                                                                    {
                                                                                        try
                                                                                        {
                                                                                            object returnValue = fieldInfo.GetValue(@object);

                                                                                            code = MarshalOps.FixupReturnValue(
                                                                                                interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                objectName, interpName, returnValue, create, dispose,
                                                                                                alias, aliasAll, toString, ref result);
                                                                                        }
                                                                                        catch (Exception e)
                                                                                        {
                                                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                                                            result = e;
                                                                                            code = ReturnCode.Error;
                                                                                        }
                                                                                    }
                                                                                    else if (argumentCount == 1)
                                                                                    {
                                                                                        try
                                                                                        {
                                                                                            code = MarshalOps.FixupArgument(
                                                                                                interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                fieldInfo.FieldType, ArgumentInfo.Create(0,
                                                                                                    fieldInfo.FieldType, newMemberName, true, false),
                                                                                                true, true, false, false, ref args[0], ref result);

                                                                                            if (code == ReturnCode.Ok)
                                                                                            {
                                                                                                fieldInfo.SetValue(
                                                                                                    @object, args[0], bindingFlags, interpreter.Binder as Binder,
                                                                                                    interpreter.CultureInfo);

                                                                                                result = String.Empty;
                                                                                            }
                                                                                        }
                                                                                        catch (Exception e)
                                                                                        {
                                                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                                                            result = e;
                                                                                            code = ReturnCode.Error;
                                                                                        }
                                                                                    }
                                                                                    else
                                                                                    {
                                                                                        result = String.Format(
                                                                                            "wrong # args for field: " +
                                                                                            "should be \"object invoke ?options? \"{0}\" \"{1}\" ?newValue?\"",
                                                                                            arguments[argumentIndex], arguments[argumentIndex + 1]);

                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                }
                                                                                else
                                                                                {
                                                                                    code = MarshalOps.FixupReturnValue(
                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                        returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                        objectName, interpName, fieldInfo, create, dispose, alias,
                                                                                        aliasAll, toString, ref result);
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "field \"{0}\" of object \"{1}\" not found",
                                                                                    arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                            break;
                                                                        }
                                                                    case MemberTypes.Method:
                                                                        {
                                                                            //
                                                                            // HACK: Cannot filter on the name here (via Type.GetMethods)?
                                                                            //
                                                                            MethodInfo[] methodInfo = objectType.GetMethods(bindingFlags);

                                                                            if (methodInfo != null)
                                                                            {
                                                                                IntList methodIndexList = null;
                                                                                ObjectArrayList argsList = null;
                                                                                ArgumentInfoList argumentInfoList = null;
                                                                                ResultList errors = null;

                                                                                //
                                                                                // NOTE: Attempt to convert the argument strings to something
                                                                                //       potentially more meaningful and find the corresponding
                                                                                //       method.
                                                                                //
                                                                                code = MarshalOps.FindMethodsAndFixupArguments(
                                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                    objectType, newObjectName, newMemberName, MemberTypes.Method,
                                                                                    bindingFlags, methodInfo, parameterTypes, args, limit, verbose,
                                                                                    @default, ref methodIndexList, ref argsList, ref argumentInfoList,
                                                                                    ref errors);

                                                                                if (code == ReturnCode.Ok)
                                                                                {
                                                                                    //
                                                                                    // NOTE: Make sure we got a valid list of method indexes.
                                                                                    //
                                                                                    if ((methodIndexList != null) && (methodIndexList.Count > 0) &&
                                                                                        (argsList != null) && (argsList.Count > 0))
                                                                                    {
                                                                                        if ((index == Index.Invalid) ||
                                                                                            ((index >= 0) && (index < methodIndexList.Count) &&
                                                                                            (index < argsList.Count)))
                                                                                        {
                                                                                            if (invoke)
                                                                                            {
                                                                                                if (!strictMember || (methodIndexList.Count == 1))
                                                                                                {
                                                                                                    //
                                                                                                    // FIXME: Select the first method that matches OR the one
                                                                                                    //        specified by the caller.  More sophisticated
                                                                                                    //        logic may need to be added here later.
                                                                                                    //
                                                                                                    int methodIndex = (index != Index.Invalid) ?
                                                                                                        methodIndexList[index] : methodIndexList[0];

                                                                                                    if (methodIndex != Index.Invalid)
                                                                                                    {
                                                                                                        try
                                                                                                        {
                                                                                                            //
                                                                                                            // NOTE: Get the arguments we are going to use to perform
                                                                                                            //       the actual method call.
                                                                                                            //
                                                                                                            args = (index != Index.Invalid) ? argsList[index] : argsList[0];

                                                                                                            if (trace)
                                                                                                                RuntimeOps.DebugTrace(String.Format(
                                                                                                                    "{0}: Execute: methodIndex = {1}, methodInfo = {2}, " +
                                                                                                                    "args = {3}{4}", GlobalState.GetCurrentThreadId(),
                                                                                                                    methodIndex, FormatOps.WrapOrNull(methodInfo[methodIndex]),
                                                                                                                    FormatOps.WrapOrNull(new StringList(args)),
                                                                                                                    Environment.NewLine), typeof(Object).Name);

                                                                                                            object returnValue = methodInfo[methodIndex].Invoke(
                                                                                                                @object, bindingFlags, interpreter.Binder as Binder, args,
                                                                                                                interpreter.CultureInfo);

                                                                                                            if (argumentInfoList != null)
                                                                                                                code = MarshalOps.FixupByRefArguments(
                                                                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                                    argumentInfoList, objectFlags,
                                                                                                                    RuntimeOps.GetInvokeOptions(aliasAll), interpName, args,
                                                                                                                    argumentFlags, strictArgs, create, dispose, alias,
                                                                                                                    aliasAll, toString, ref result);

                                                                                                            if (code == ReturnCode.Ok)
                                                                                                                code = MarshalOps.FixupReturnValue(
                                                                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                                    returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                                    objectName, interpName, returnValue, create, dispose,
                                                                                                                    alias, aliasAll, toString, ref result);
                                                                                                        }
                                                                                                        catch (Exception e)
                                                                                                        {
                                                                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                                                                            result = e;
                                                                                                            code = ReturnCode.Error;
                                                                                                        }
                                                                                                    }
                                                                                                    else
                                                                                                    {
                                                                                                        result = String.Format(
                                                                                                            "method \"{0}\" of object \"{1}\" not found",
                                                                                                            arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                                        code = ReturnCode.Error;
                                                                                                    }
                                                                                                }
                                                                                                else
                                                                                                {
                                                                                                    result = String.Format(
                                                                                                        "matched {0} method overloads of \"{1}\" on type \"{2}\", need exactly 1",
                                                                                                         methodIndexList.Count, arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                                    code = ReturnCode.Error;
                                                                                                }
                                                                                            }
                                                                                            else
                                                                                            {
                                                                                                MethodInfoList methodInfoList = new MethodInfoList();

                                                                                                if (index != Index.Invalid)
                                                                                                    methodInfoList.Add(methodInfo[methodIndexList[index]]);
                                                                                                else
                                                                                                    foreach (int methodIndex in methodIndexList)
                                                                                                        methodInfoList.Add(methodInfo[methodIndex]);

                                                                                                code = MarshalOps.FixupReturnValue(
                                                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                    returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                    objectName, interpName, methodInfoList, create, dispose,
                                                                                                    alias, aliasAll, toString, ref result);
                                                                                            }
                                                                                        }
                                                                                        else
                                                                                        {
                                                                                            result = String.Format(
                                                                                                "method \"{0}\" of object \"{1}\" not found, " +
                                                                                                "invalid method index {2}, must be {3}",
                                                                                                arguments[argumentIndex + 1], arguments[argumentIndex],
                                                                                                index, FormatOps.BetweenOrExact(0, methodIndexList.Count - 1));

                                                                                            code = ReturnCode.Error;
                                                                                        }
                                                                                    }
                                                                                    else
                                                                                    {
                                                                                        result = String.Format(
                                                                                            "method \"{0}\" of object \"{1}\" not found",
                                                                                            arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                }
                                                                                else
                                                                                {
                                                                                    result = errors;
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "type \"{0}\" has no methods matching member " +
                                                                                    "types \"{1}\" and flags \"{2}\"",
                                                                                    objectType.FullName, memberTypes, bindingFlags);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                            break;
                                                                        }
                                                                    case MemberTypes.Property:
                                                                        {
                                                                            //
                                                                            // HACK: Cannot filter on the name here (via Type.GetProperties)?
                                                                            //
                                                                            PropertyInfo[] propertyInfo = objectType.GetProperties(bindingFlags);

                                                                            if (propertyInfo != null)
                                                                            {
                                                                                MethodInfo[] methodInfo = null;

                                                                                //
                                                                                // NOTE: Ok, now we need to get all the methods that access
                                                                                //       the properties (get and set).
                                                                                //
                                                                                code = MarshalOps.GetMethodInfoFromPropertyInfo(
                                                                                    propertyInfo, bindingFlags, ref methodInfo, ref result);

                                                                                if (code == ReturnCode.Ok)
                                                                                {
                                                                                    IntList methodIndexList = null;
                                                                                    ObjectArrayList argsList = null;
                                                                                    ArgumentInfoList argumentInfoList = null;
                                                                                    ResultList errors = null;

                                                                                    //
                                                                                    // NOTE: Attempt to convert the argument strings to something
                                                                                    //       potentially more meaningful and find the corresponding
                                                                                    //       method.
                                                                                    //
                                                                                    code = MarshalOps.FindMethodsAndFixupArguments(
                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                        objectType, newObjectName, newMemberName, MemberTypes.Property,
                                                                                        bindingFlags, methodInfo, parameterTypes, args, limit, verbose,
                                                                                        @default, ref methodIndexList, ref argsList, ref argumentInfoList,
                                                                                        ref errors);

                                                                                    if (code == ReturnCode.Ok)
                                                                                    {
                                                                                        if ((methodIndexList != null) && (methodIndexList.Count > 0) &&
                                                                                            (argsList != null) && (argsList.Count > 0))
                                                                                        {
                                                                                            if ((index == Index.Invalid) ||
                                                                                                ((index >= 0) && (index < methodIndexList.Count) &&
                                                                                                (index < argsList.Count)))
                                                                                            {
                                                                                                if (invoke)
                                                                                                {
                                                                                                    if (!strictMember || (methodIndexList.Count == 1))
                                                                                                    {
                                                                                                        //
                                                                                                        // FIXME: Select the first method that matches OR the one
                                                                                                        //        specified by the caller.  More sophisticated
                                                                                                        //        logic may need to be added here later.
                                                                                                        //
                                                                                                        int methodIndex = (index != Index.Invalid) ?
                                                                                                            methodIndexList[index] : methodIndexList[0];

                                                                                                        if (methodIndex != Index.Invalid)
                                                                                                        {
                                                                                                            try
                                                                                                            {
                                                                                                                //
                                                                                                                // NOTE: Get the arguments we are going to use to perform
                                                                                                                //       the actual method call.
                                                                                                                //
                                                                                                                args = (index != Index.Invalid) ? argsList[index] : argsList[0];

                                                                                                                if (trace)
                                                                                                                    RuntimeOps.DebugTrace(String.Format(
                                                                                                                        "{0}: Execute: methodIndex = {1}, methodInfo = {2}, " +
                                                                                                                        "args = {3}{4}", GlobalState.GetCurrentThreadId(),
                                                                                                                        methodIndex, FormatOps.WrapOrNull(methodInfo[methodIndex]),
                                                                                                                        FormatOps.WrapOrNull(new StringList(args)),
                                                                                                                        Environment.NewLine), typeof(Object).Name);

                                                                                                                object returnValue = methodInfo[methodIndex].Invoke(
                                                                                                                    @object, bindingFlags, interpreter.Binder as Binder, args,
                                                                                                                    interpreter.CultureInfo);

                                                                                                                if (argumentInfoList != null)
                                                                                                                    code = MarshalOps.FixupByRefArguments(
                                                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                                        argumentInfoList, objectFlags,
                                                                                                                        RuntimeOps.GetInvokeOptions(aliasAll), interpName, args,
                                                                                                                        argumentFlags, strictArgs, create, dispose, alias,
                                                                                                                        aliasAll, toString, ref result);

                                                                                                                if (code == ReturnCode.Ok)
                                                                                                                    code = MarshalOps.FixupReturnValue(
                                                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                                        returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                                        objectName, interpName, returnValue, create, dispose,
                                                                                                                        alias, aliasAll, toString, ref result);
                                                                                                            }
                                                                                                            catch (Exception e)
                                                                                                            {
                                                                                                                Engine.SetExceptionErrorCode(interpreter, e);

                                                                                                                result = e;
                                                                                                                code = ReturnCode.Error;
                                                                                                            }
                                                                                                        }
                                                                                                        else
                                                                                                        {
                                                                                                            result = String.Format(
                                                                                                                "property \"{0}\" of object \"{1}\" not found",
                                                                                                                arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                                            code = ReturnCode.Error;
                                                                                                        }
                                                                                                    }
                                                                                                    else
                                                                                                    {
                                                                                                        result = String.Format(
                                                                                                            "matched {0} property overloads of \"{1}\" on type \"{2}\", need exactly 1",
                                                                                                             methodIndexList.Count, arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                                        code = ReturnCode.Error;
                                                                                                    }
                                                                                                }
                                                                                                else
                                                                                                {
                                                                                                    MethodInfoList methodInfoList = new MethodInfoList();

                                                                                                    if (index != Index.Invalid)
                                                                                                        methodInfoList.Add(methodInfo[methodIndexList[index]]);
                                                                                                    else
                                                                                                        foreach (int methodIndex in methodIndexList)
                                                                                                            methodInfoList.Add(methodInfo[methodIndex]);

                                                                                                    code = MarshalOps.FixupReturnValue(
                                                                                                        interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                                                        returnType, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                                                        objectName, interpName, methodInfoList, create, dispose,
                                                                                                        alias, aliasAll, toString, ref result);
                                                                                                }
                                                                                            }
                                                                                            else
                                                                                            {
                                                                                                result = String.Format(
                                                                                                    "property \"{0}\" of object \"{1}\" not found, " +
                                                                                                    "invalid method index {2}, must be {3}",
                                                                                                    arguments[argumentIndex + 1], arguments[argumentIndex],
                                                                                                    index, FormatOps.BetweenOrExact(0, methodIndexList.Count - 1));

                                                                                                code = ReturnCode.Error;
                                                                                            }
                                                                                        }
                                                                                        else
                                                                                        {
                                                                                            result = String.Format(
                                                                                                "property \"{0}\" of object \"{1}\" not found",
                                                                                                arguments[argumentIndex + 1], arguments[argumentIndex]);

                                                                                            code = ReturnCode.Error;
                                                                                        }
                                                                                    }
                                                                                    else
                                                                                    {
                                                                                        result = errors;
                                                                                    }
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "type \"{0}\" has no properties matching \"{1}\"",
                                                                                    objectType.FullName, bindingFlags);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                            break;
                                                                        }
                                                                    default:
                                                                        {
                                                                            result = String.Format(
                                                                                "unsupported member type \"{0}\"",
                                                                                memberInfo[0].MemberType);

                                                                            code = ReturnCode.Error;
                                                                            break;
                                                                        }
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "type \"{0}\" member \"{1}\" matching member " +
                                                                    "types \"{2}\" and flags \"{3}\" not found",
                                                                    objectType.FullName, arguments[argumentIndex + 1],
                                                                    memberTypes, bindingFlags);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object invoke ?options? object member ?arg ...?\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object invoke ?options? object member ?arg ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "invokeall":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = RuntimeOps.GetInvokeAllOptions();
                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false,
                                                ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) < arguments.Count))
                                                {
                                                    bool invoke = false;

                                                    if (options.IsPresent("-invoke"))
                                                        invoke = true;

                                                    //
                                                    // NOTE: If we are supposed to simply use [object invoke] for this
                                                    //       call instead, do that now.
                                                    //
                                                    if (invoke)
                                                    {
                                                        //
                                                        // NOTE: Copy our current argument list verbatim.
                                                        //
                                                        ArgumentList newArguments = new ArgumentList(arguments);

                                                        //
                                                        // NOTE: Replace the sub-command name with "invoke".
                                                        //
                                                        newArguments[1] = "invoke";

                                                        //
                                                        // NOTE: Invoke this method recursively with the modified
                                                        //       argument list.
                                                        //
                                                        return Execute(interpreter, clientData, newArguments, ref result);
                                                    }

                                                    bool keepResults = false;

                                                    if (options.IsPresent("-keepresults"))
                                                        keepResults = true;

                                                    bool noComplain = false;

                                                    if (options.IsPresent("-nocomplain"))
                                                        noComplain = true;

                                                    //
                                                    // NOTE: If necessary, mask-off the options that are not supported by
                                                    //       [object invoke].
                                                    //
                                                    if (keepResults)
                                                        options.SetPresent("-keepresults", false, Index.Invalid, null);

                                                    if (noComplain)
                                                        options.SetPresent("-nocomplain", false, Index.Invalid, null);

                                                    //
                                                    // NOTE: Save the object name because we need it every time through
                                                    //       the loop.
                                                    //
                                                    string objectName = arguments[argumentIndex];

                                                    //
                                                    // NOTE: This will contain the number of invocation errors.
                                                    //
                                                    int errorCount = 0;

                                                    //
                                                    // NOTE: If we are keeping track of all the results, create the
                                                    //       necessary list now.
                                                    //
                                                    ResultList newResults = keepResults ? new ResultList() : null;

                                                    //
                                                    // NOTE: Advance to the first invocation list element and keep
                                                    //       going until there are no more to be processed (unless
                                                    //       we break out early).
                                                    //
                                                    for (argumentIndex++; argumentIndex < arguments.Count; argumentIndex++)
                                                    {
                                                        //
                                                        // NOTE: First, parse the invocation list element.  It must
                                                        //       be a list consisting of a member name followed by
                                                        //       the arguments to that member, if any.
                                                        //
                                                        StringList list = null;
                                                        Result newResult = null;

                                                        code = Parser.SplitList(
                                                            interpreter, arguments[argumentIndex], 0, Length.Invalid,
                                                            false, true, ref list, ref newResult);

                                                        if ((code != ReturnCode.Ok) || (list.Count == 0))
                                                        {
                                                            if (code == ReturnCode.Ok)
                                                            {
                                                                newResult = "invocation list element missing member name";
                                                                code = ReturnCode.Error;
                                                            }

                                                            if (noComplain)
                                                                noComplain = false; /* NOTE: Not an invocation error. */

                                                            goto error;
                                                        }

                                                        //
                                                        // NOTE: Next, grab the first two arguments, which are currently
                                                        //       always the same (i.e. [object invoke]) and append them to
                                                        //       the new argument list.
                                                        //
                                                        ArgumentList newArguments = null;

                                                        code = RuntimeOps.GetObjectAliasArguments(false, ref newArguments,
                                                            ref newResult);

                                                        if (code != ReturnCode.Ok)
                                                        {
                                                            if (noComplain)
                                                                noComplain = false; /* NOTE: Not an invocation error. */

                                                            goto error;
                                                        }

                                                        //
                                                        // NOTE: Next, add all the options used to invoke this command with
                                                        //       the exception of the ones we [may] have masked off earlier
                                                        //       (i.e. the ones supported by us and not [object invoke]).
                                                        //
                                                        code = options.ToArgumentList(ref newArguments, ref newResult);

                                                        if (code != ReturnCode.Ok)
                                                        {
                                                            if (noComplain)
                                                                noComplain = false; /* NOTE: Not an invocation error. */

                                                            goto error;
                                                        }

                                                        //
                                                        // NOTE: Next, add the object name, which is currently always the
                                                        //       same, to the new argument list.
                                                        //
                                                        newArguments.Add(objectName);

                                                        //
                                                        // NOTE: Next, add the member name and its associated arguments, if
                                                        //       any.
                                                        //
                                                        newArguments.AddRange(list);

                                                        //
                                                        // NOTE: Finally, invoke this method again using the new arguments
                                                        //       and a new result.
                                                        //
                                                        /* RECURSIVE */
                                                        code = Execute(interpreter, clientData, newArguments, ref newResult);

                                                    error:
                                                        //
                                                        // NOTE: If we are keeping track of results, add the return code and
                                                        //       result to the result list now.  We need to do this whether
                                                        //       or not we are ignoring invocation errors.
                                                        //
                                                        if (keepResults && (newResults != null))
                                                        {
                                                            newResults.Add(code);
                                                            newResults.Add(newResult);
                                                        }

                                                        //
                                                        // NOTE: Did the invocation raise some kind of error?  If so, we need
                                                        //       to determine how to handle it.
                                                        //
                                                        if (code != ReturnCode.Ok)
                                                        {
                                                            //
                                                            // NOTE: If we are ignoring invocation errors, just increase the
                                                            //       error count and then pretend like nothing bad happened;
                                                            //       otherwise, we will bail out after setting the overall
                                                            //       result.
                                                            //
                                                            if (noComplain)
                                                            {
                                                                errorCount++;
                                                                code = ReturnCode.Ok;
                                                            }
                                                            else
                                                            {
                                                                //
                                                                // NOTE: If we have been keeping track of results the whole
                                                                //       time, do not discard them now; otherwise, just return
                                                                //       the error that is causing us to abort our processing.
                                                                //
                                                                if (keepResults && (newResults != null))
                                                                    result = newResults;
                                                                else
                                                                    result = newResult;

                                                                break;
                                                            }
                                                        }
                                                    }

                                                    //
                                                    // NOTE: Was the overall result success?
                                                    //
                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        //
                                                        // NOTE: If we were keeping track of results the whole time, return
                                                        //       them all as one big list; otherwise, return the literal
                                                        //       string "Ok" if zero errors were encountered OR the literal
                                                        //       string "Error" if there was at least one error.
                                                        //
                                                        if (keepResults && (newResults != null))
                                                            result = newResults;
                                                        else
                                                            result = StringList.MakeList((errorCount == 0) ?
                                                                ReturnCode.Ok : ReturnCode.Error, errorCount);
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object invokeall ?options? object memberAndArgs ?memberAndArgs ...?\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object invokeall ?options? object memberAndArgs ?memberAndArgs ...?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "list":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            if (interpreter.objects != null)
                                            {
                                                string pattern = null;

                                                if (arguments.Count == 3)
                                                    pattern = arguments[2];

                                                StringList list = new StringList();

                                                foreach (KeyValuePair<string, _Wrappers._Object> pair in interpreter.objects)
                                                {
                                                    if ((pattern == null) ||
                                                        Parser.StringMatch(interpreter, pair.Key, 0, pattern, 0, false))
                                                    {
                                                        list.Add(pair.Key);
                                                    }
                                                }

                                                result = list;
                                            }
                                            else
                                            {
                                                result = String.Empty;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object list ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "load":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-create", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-import", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-importnonpublic", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-importmode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-importpattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-importnocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declare", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declarenonpublic", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-declaremode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-declarepattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declarenocase", null),
                                                new Option(typeof(LoadType), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-loadtype", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default | ObjectFlags.Assembly)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) == arguments.Count))
                                                {
                                                    //
                                                    // NOTE: This defaults to false because we know that Assembly
                                                    //       is not a primitive type and hence that it will not
                                                    //       be automatically converted by FixupReturnValue.
                                                    //
                                                    bool create = false;

                                                    if (options.IsPresent("-create"))
                                                        create = true;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    //
                                                    // NOTE: Do we want to create a command alias for any newly created
                                                    //       objects?
                                                    //
                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    //
                                                    // NOTE: Do we want to import the namespaces from the assembly that
                                                    //       contain public types?
                                                    //
                                                    bool import = false;

                                                    if (options.IsPresent("-import"))
                                                        import = true;

                                                    //
                                                    // NOTE: Do we want to import the namespaces for non-public types as
                                                    //       well?
                                                    //
                                                    bool importNonPublic = false;

                                                    if (options.IsPresent("-importnonpublic"))
                                                        importNonPublic = true;

                                                    //
                                                    // NOTE: Do we want to import the declare from the assembly?  This
                                                    //       is normally only used for COM Interop assemblies.
                                                    //
                                                    bool declare = false;

                                                    if (options.IsPresent("-declare"))
                                                        declare = true;

                                                    //
                                                    // NOTE: Do we want to declare non-public interfaces as well?
                                                    //
                                                    bool declareNonPublic = false;

                                                    if (options.IsPresent("-declarenonpublic"))
                                                        declareNonPublic = true;

                                                    //
                                                    // NOTE: What match mode are we going to use to restrict the imported
                                                    //       namespaces, if any?
                                                    //
                                                    Variant value = null;
                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    MatchMode importMode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-importmode", ref value))
                                                        importMode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What match mode are we going to use to restrict the declared
                                                    //       interfaces, if any?
                                                    //
                                                    MatchMode declareMode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-declaremode", ref value))
                                                        declareMode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict the imported
                                                    //       namespaces, if any?
                                                    //
                                                    string importPattern = null;

                                                    if (options.IsPresent("-importpattern", ref value))
                                                        importPattern = value.ToString();

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict the declared
                                                    //       interfaces, if any?
                                                    //
                                                    string declarePattern = null;

                                                    if (options.IsPresent("-declarepattern", ref value))
                                                        declarePattern = value.ToString();

                                                    //
                                                    // NOTE: Which Tcl interpreter do we want a command alias created
                                                    //       in, if any?
                                                    //
                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default | ObjectFlags.Assembly;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    //
                                                    // NOTE: For the purposes of matching namespace names, do we care
                                                    //       about case?
                                                    //
                                                    bool importNoCase = false;

                                                    if (options.IsPresent("-importnocase"))
                                                        importNoCase = true;

                                                    //
                                                    // NOTE: For the purposes of matching interface names, do we care
                                                    //       about case?
                                                    //
                                                    bool declareNoCase = false;

                                                    if (options.IsPresent("-declarenocase"))
                                                        declareNoCase = true;

                                                    //
                                                    // NOTE: Forcibly convert the object to a string and discard it?
                                                    //
                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    LoadType loadType = LoadType.Default;

                                                    if (options.IsPresent("-loadtype", ref value))
                                                        loadType = (LoadType)value.Value;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((importPattern != null) && (importMode == MatchMode.None))
                                                        importMode = StringOps.DefaultObjectMatchMode;

                                                    if ((declarePattern != null) && (declareMode == MatchMode.None))
                                                        declareMode = StringOps.DefaultObjectMatchMode;

                                                    //
                                                    // NOTE: We are going to modify the interpreter state, make
                                                    //       sure it is not set to read-only.
                                                    //
                                                    if (!interpreter.ReadOnly)
                                                    {
                                                        Assembly assembly = null;

                                                        try
                                                        {
                                                            switch (loadType)
                                                            {
                                                                case LoadType.PartialName:
                                                                    {
#if MONO_BUILD
#pragma warning disable 612
#endif
                                                                        assembly = Assembly.LoadWithPartialName(
                                                                            arguments[argumentIndex]);
#if MONO_BUILD
#pragma warning restore 612
#endif

                                                                        if (assembly == null)
                                                                        {
                                                                            result = String.Format(
                                                                                "could not load assembly based on partial name \"{0}\"",
                                                                                arguments[argumentIndex]);

                                                                            code = ReturnCode.Error;
                                                                        }
                                                                        break;
                                                                    }
                                                                case LoadType.FullName:
                                                                    {
                                                                        assembly = Assembly.Load(
                                                                            arguments[argumentIndex]);
                                                                        break;
                                                                    }
                                                                case LoadType.File:
                                                                    {
                                                                        assembly = Assembly.LoadFrom(
                                                                            arguments[argumentIndex]);
                                                                        break;
                                                                    }
                                                                case LoadType.Bytes:
                                                                    {
                                                                        assembly = Assembly.Load(
                                                                            Convert.FromBase64String(
                                                                                arguments[argumentIndex]));
                                                                        break;
                                                                    }
                                                                case LoadType.Stream:
                                                                    {
                                                                        IObject @object = null;

                                                                        code = interpreter.GetObject(
                                                                            arguments[argumentIndex], true,
                                                                            true, ref @object, ref result);

                                                                        if (code == ReturnCode.Ok)
                                                                        {
                                                                            Stream stream = @object.Value as Stream;

                                                                            if (stream != null)
                                                                            {
                                                                                if (stream.CanRead && stream.CanSeek)
                                                                                {
                                                                                    long length = stream.Length;

                                                                                    if (length <= int.MaxValue)
                                                                                    {
                                                                                        int intLength = (int)length;
                                                                                        byte[] bytes = new byte[intLength];

                                                                                        stream.Read(bytes, 0, intLength);
                                                                                        assembly = Assembly.Load(bytes);
                                                                                    }
                                                                                    else
                                                                                    {
                                                                                        result = String.Format(
                                                                                            "stream for object \"{0}\" too large",
                                                                                            arguments[argumentIndex]);

                                                                                        code = ReturnCode.Error;
                                                                                    }
                                                                                }
                                                                                else
                                                                                {
                                                                                    result = String.Format(
                                                                                        "stream for object \"{0}\" must " +
                                                                                        "support seeking and reading",
                                                                                        arguments[argumentIndex]);

                                                                                    code = ReturnCode.Error;
                                                                                }
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "invalid stream for object \"{0}\"",
                                                                                    arguments[argumentIndex]);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                        }
                                                                        break;
                                                                    }
                                                                default:
                                                                    {
                                                                        result = String.Format(
                                                                            "unsupported assembly load type \"{0}\"",
                                                                            loadType);

                                                                        code = ReturnCode.Error;
                                                                        break;
                                                                    }
                                                            }
                                                        }
                                                        catch (Exception e)
                                                        {
                                                            Engine.SetExceptionErrorCode(interpreter, e);

                                                            result = e;
                                                            code = ReturnCode.Error;
                                                        }

                                                        if (code == ReturnCode.Ok)
                                                            code = MarshalOps.FixupReturnValue(
                                                                interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                interpName, objectName, assembly, create, dispose, alias,
                                                                aliasAll, toString, ref result);

                                                        if ((code == ReturnCode.Ok) && import)
                                                            code = interpreter.AddObjectNamespaces(assembly, importNonPublic,
                                                                importMode, importPattern, importNoCase, ref result);

                                                        if ((code == ReturnCode.Ok) && declare)
                                                            code = interpreter.AddObjectInterfaces(assembly, declareNonPublic,
                                                                declareMode, declarePattern, declareNoCase,
                                                                ref result);
                                                    }
                                                    else
                                                    {
                                                        result = "interpreter is read-only";
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object load ?options? assembly\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object load ?options? assembly\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "members":
                                    {
                                        if (arguments.Count >= 3)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-mode", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-pattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-attributes", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-signatures", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-qualified", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nameonly", null),
                                                new Option(typeof(MemberTypes), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-membertypes",
                                                    new Variant(MarshalOps.DefaultMemberTypes)),
                                                new Option(typeof(BindingFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-flags",
                                                    new Variant(MarshalOps.DefaultBindingFlags)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, false, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) == arguments.Count))
                                                {
                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    //
                                                    // NOTE: Prevent any magical type searches (i.e. use their specified
                                                    //       type string verbatim)?
                                                    //
                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    //
                                                    // NOTE: Return all Type exception information (this can be very
                                                    //       costly for performance).
                                                    //
                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    //
                                                    // NOTE: Return the associated attributes for all returned members?
                                                    //
                                                    bool attributes = false;

                                                    if (options.IsPresent("-attributes"))
                                                        attributes = true;

                                                    //
                                                    // NOTE: Return the method signatures for properties and methods?
                                                    //
                                                    bool signatures = false;

                                                    if (options.IsPresent("-signatures"))
                                                        signatures = true;

                                                    //
                                                    // NOTE: Use fully-qualified type names for properties and methods?
                                                    //
                                                    bool qualified = false;

                                                    if (options.IsPresent("-qualified"))
                                                        qualified = true;

                                                    //
                                                    // NOTE: Only return the names of the members, not the member types.
                                                    //
                                                    bool nameOnly = false;

                                                    if (options.IsPresent("-nameonly"))
                                                        nameOnly = true;

                                                    //
                                                    // NOTE: What match mode are we going to use to restrict the members,
                                                    //       if any?
                                                    //
                                                    Variant value = null;
                                                    MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-mode", ref value))
                                                        mode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict the members,
                                                    //       if any?
                                                    //
                                                    string pattern = null;

                                                    if (options.IsPresent("-pattern", ref value))
                                                        pattern = value.ToString();

                                                    MemberTypes memberTypes = MarshalOps.DefaultMemberTypes;

                                                    if (options.IsPresent("-membertypes", ref value))
                                                        memberTypes = (MemberTypes)value.Value;

                                                    BindingFlags bindingFlags = MarshalOps.DefaultBindingFlags;

                                                    if (options.IsPresent("-flags", ref value))
                                                        bindingFlags = (BindingFlags)value.Value;

                                                    //
                                                    // NOTE: Now check for and use the -nocase value.  It is also
                                                    //       important to note here that a specifying the flags
                                                    //       does not override this setting.
                                                    //
                                                    if (noCase)
                                                        bindingFlags |= BindingFlags.IgnoreCase;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((pattern != null) && (mode == MatchMode.None))
                                                        mode = StringOps.DefaultObjectMatchMode;

                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    if (type == null)
                                                    {
                                                        IObject @object = null;

                                                        if (interpreter.GetObject(
                                                                arguments[argumentIndex], true, true,
                                                                ref @object) == ReturnCode.Ok)
                                                        {
                                                            if (@object.Value != null)
                                                            {
                                                                type = @object.Value.GetType();
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "invalid value for object \"{0}\"",
                                                                    arguments[argumentIndex]);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            ResultList errors = null;

                                                            code = Value.GetType(interpreter, arguments[argumentIndex],
                                                                interpreter.GetAppDomain(), strictType, verbose, noCase,
                                                                interpreter.CultureInfo, ref type, ref errors);

                                                            if (code != ReturnCode.Ok)
                                                            {
                                                                errors.Insert(0, String.Format(
                                                                    "object or type \"{0}\" not found",
                                                                    arguments[argumentIndex]));

                                                                result = errors;
                                                            }
                                                        }
                                                    }

                                                    //
                                                    // NOTE: Did the type lookup above succeed?
                                                    //
                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        MemberInfo[] memberInfo = type.GetMembers(bindingFlags);

                                                        if ((memberInfo != null) && (memberInfo.Length > 0))
                                                        {
                                                            StringDictionary dictionary = new StringDictionary();

                                                            for (int index = 0; index < memberInfo.Length; index++)
                                                            {
                                                                if (FlagOps.HasFlags(memberTypes, memberInfo[index].MemberType, true))
                                                                {
                                                                    StringList list;

                                                                    if (nameOnly)
                                                                        list = new StringList(
                                                                            memberInfo[index].Name);
                                                                    else
                                                                        list = new StringList(
                                                                            "memberType",
                                                                            memberInfo[index].MemberType.ToString(),
                                                                            "memberName",
                                                                            memberInfo[index].Name);

                                                                    if (attributes)
                                                                    {
                                                                        EventAttributes? eventAttributes =
                                                                            MarshalOps.GetEventAttributes(memberInfo[index]);

                                                                        if (eventAttributes != null)
                                                                        {
                                                                            list.Add("eventAttributes");
                                                                            list.Add(eventAttributes.ToString());
                                                                        }

                                                                        FieldAttributes? fieldAttributes =
                                                                            MarshalOps.GetFieldAttributes(memberInfo[index]);

                                                                        if (fieldAttributes != null)
                                                                        {
                                                                            list.Add("fieldAttributes");
                                                                            list.Add(fieldAttributes.ToString());
                                                                        }

                                                                        MethodAttributes? methodAttributes =
                                                                            MarshalOps.GetMethodAttributes(memberInfo[index]);

                                                                        if (methodAttributes != null)
                                                                        {
                                                                            list.Add("methodAttributes");
                                                                            list.Add(methodAttributes.ToString());
                                                                        }

                                                                        PropertyAttributes? propertyAttributes =
                                                                            MarshalOps.GetPropertyAttributes(memberInfo[index]);

                                                                        if (propertyAttributes != null)
                                                                        {
                                                                            list.Add("propertyAttributes");
                                                                            list.Add(propertyAttributes.ToString());
                                                                        }

                                                                        TypeAttributes? typeAttributes =
                                                                            MarshalOps.GetTypeAttributes(memberInfo[index]);

                                                                        if (typeAttributes != null)
                                                                        {
                                                                            list.Add("typeAttributes");
                                                                            list.Add(typeAttributes.ToString());
                                                                        }
                                                                    }

                                                                    //
                                                                    // NOTE: Check if we want all the method signatures.  This
                                                                    //       option does NOT apply to fields or types because
                                                                    //       they never have a MethodInfo associated with them.
                                                                    //
                                                                    if (signatures)
                                                                    {
                                                                        if (!(memberInfo[index] is FieldInfo) &&
                                                                            !(memberInfo[index] is Type))
                                                                        {
                                                                            MethodBase[] methodBase = null;

                                                                            code = MarshalOps.GetMethodBaseFromMemberInfo(
                                                                                new MemberInfo[] { memberInfo[index] },
                                                                                bindingFlags, ref methodBase, ref result);

                                                                            if (code == ReturnCode.Ok)
                                                                            {
                                                                                for (int index2 = 0; index2 < methodBase.Length; index2++)
                                                                                {
                                                                                    TypeList returnType = null;

                                                                                    code = MarshalOps.GetTypeListFromParameterInfo(
                                                                                        new ParameterInfo[] {
                                                                                        MarshalOps.GetReturnParameterInfo(methodBase[index2]) },
                                                                                        true, ref returnType, ref result);

                                                                                    if (code != ReturnCode.Ok)
                                                                                        break;

                                                                                    TypeList parameterTypes = null;

                                                                                    code = MarshalOps.GetTypeListFromParameterInfo(
                                                                                        methodBase[index2].GetParameters(), false,
                                                                                        ref parameterTypes, ref result);

                                                                                    if (code != ReturnCode.Ok)
                                                                                        break;

                                                                                    list.AddRange(new StringList(
                                                                                        "methodType",
                                                                                        //
                                                                                        // NOTE: This should always be "Method"; however, we
                                                                                        //       do need to include it for script comparison
                                                                                        //       purposes and we do not want to hard-code it.
                                                                                        //
                                                                                        methodBase[index2].MemberType.ToString(),
                                                                                        "methodName",
                                                                                        methodBase[index2].Name,
                                                                                        "callingConvention",
                                                                                        methodBase[index2].CallingConvention.ToString(),
                                                                                        "returnType",
                                                                                        returnType.ToString(null, false, qualified),
                                                                                        "parameterTypes",
                                                                                        parameterTypes.ToString(null, false, qualified)));
                                                                                }
                                                                            }
                                                                        }
                                                                        else if (memberInfo[index] is FieldInfo)
                                                                        {
                                                                            FieldInfo fieldInfo = (FieldInfo)memberInfo[index];

                                                                            if (fieldInfo != null)
                                                                            {
                                                                                TypeList fieldType = new TypeList(
                                                                                    new Type[] { fieldInfo.FieldType });

                                                                                list.AddRange(new StringList(
                                                                                    "fieldType",
                                                                                    fieldType.ToString(null, false, qualified)));
                                                                            }
                                                                            else
                                                                            {
                                                                                result = String.Format(
                                                                                    "invalid field info for member \"{0}\"",
                                                                                    memberInfo[index].Name);

                                                                                code = ReturnCode.Error;
                                                                            }
                                                                        }
                                                                    }

                                                                    if (code != ReturnCode.Ok)
                                                                        break;

                                                                    if (list != null)
                                                                    {
                                                                        if ((mode == MatchMode.None) ||
                                                                            StringOps.Match(mode, list.ToString(), pattern, noCase))
                                                                        {
                                                                            //
                                                                            // BUGFIX: Avoid adding duplicate members (this could
                                                                            //         previously happen if a method had multiple
                                                                            //         overloads and the "-signatures" option was not
                                                                            //         specified).
                                                                            //
                                                                            string key = list.ToString();

                                                                            if (!dictionary.ContainsKey(key))
                                                                                dictionary.Add(key, null);
                                                                        }
                                                                    }
                                                                }
                                                            }

                                                            if (code == ReturnCode.Ok)
                                                                result = dictionary.ToString();
                                                        }
                                                        else
                                                        {
                                                            result = String.Empty;
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object members ?options? object\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object members ?options? object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "namespaces":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            lock (interpreter.SyncRoot)
                                            {
                                                string pattern = null;

                                                if (arguments.Count == 3)
                                                    pattern = arguments[2];

                                                StringDictionary dictionary = interpreter.ObjectNamespaces;

                                                if (dictionary != null)
                                                    result = dictionary.KeysAndValuesToString(pattern, false);
                                                else
                                                    result = String.Empty;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object namespaces ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "refcount":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            IObject @object = null;

                                            code = interpreter.GetObject(arguments[2], true, true, ref @object, ref result);

                                            if (code == ReturnCode.Ok)
                                                result = @object.ReferenceCount;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object refcount object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "removecallback":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            code = interpreter.RemoveCallback(arguments[2], clientData, ref result);

                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object removecallback name\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "resolve":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            code = MarshalOps.ResolveAssembly(arguments[2], ref result);
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object resolve assembly\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "strongname":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            IObject @object = null;

                                            code = interpreter.GetObject(arguments[2], true, true, ref @object, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                Assembly assembly = (@object != null) ? @object.Value as Assembly : null;
                                                StrongName strongName = null;

                                                code = AssemblyOps.GetStrongName(assembly, ref strongName, ref result);

                                                if (code == ReturnCode.Ok)
                                                    result = FormatOps.StrongName(assembly, strongName, true);
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object strongname assembly\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "type":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-typemode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-typepattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-typenocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 1) < arguments.Count))
                                                {
                                                    //
                                                    // NOTE: What match mode are we going to use to restrict, if any?
                                                    //
                                                    Variant value = null;
                                                    MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-typemode", ref value))
                                                        mode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict, if any?
                                                    //
                                                    string pattern = null;

                                                    if (options.IsPresent("-typepattern", ref value))
                                                        pattern = value.ToString();

                                                    //
                                                    // NOTE: For the purposes of matching names, do we care about
                                                    //       case?
                                                    //
                                                    bool noCase = false;

                                                    if (options.IsPresent("-typenocase"))
                                                        noCase = true;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((pattern != null) && (mode == MatchMode.None))
                                                        mode = StringOps.DefaultObjectMatchMode;

                                                    //
                                                    // NOTE: Figure out which ones they want to add.
                                                    //
                                                    StringList list = new StringList();

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        //
                                                        // NOTE: Any specific names they want to add?
                                                        //
                                                        if (argumentIndex != Index.Invalid)
                                                            list.Add(ArgumentList.GetRange(
                                                                arguments, argumentIndex, Index.Invalid));

                                                        //
                                                        // NOTE: Ok, add the specified names now.
                                                        //
                                                        code = interpreter.AddObjectTypes(
                                                            list, mode, pattern, noCase, ref result);

                                                        if (code == ReturnCode.Ok)
                                                            result = String.Empty;
                                                    }
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"object type ?options? ?fromName toName...? fromName toName\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object type ?options? ?fromName toName...? fromName toName\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "types":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            lock (interpreter.SyncRoot)
                                            {
                                                string pattern = null;

                                                if (arguments.Count == 3)
                                                    pattern = arguments[2];

                                                StringDictionary list = interpreter.ObjectTypes;

                                                if (list != null)
                                                    result = list.KeysAndValuesToString(pattern, false);
                                                else
                                                    result = String.Empty;
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object types ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "unalias":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            code = interpreter.RemoveAliasAndCommand(arguments[2], clientData, ref result);

                                            if (code == ReturnCode.Ok)
                                                result = String.Empty;
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object unalias object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "undeclare":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-bycontainer", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-declaremode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-declarepattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-declarenocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex == Index.Invalid)
                                                {
                                                    //
                                                    // NOTE: What match mode are we going to use to restrict the declared
                                                    //       interfaces, if any?
                                                    //
                                                    Variant value = null;
                                                    MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-declaremode", ref value))
                                                        mode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict the imported
                                                    //       interfaces, if any?
                                                    //
                                                    string pattern = null;

                                                    if (options.IsPresent("-declarepattern", ref value))
                                                        pattern = value.ToString();

                                                    //
                                                    // NOTE: For the purposes of matching interface names, do we care
                                                    //       about case?
                                                    //
                                                    bool noCase = false;

                                                    if (options.IsPresent("-declarenocase"))
                                                        noCase = true;

                                                    bool values = false;

                                                    if (options.IsPresent("-bycontainer"))
                                                        values = true;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((pattern != null) && (mode == MatchMode.None))
                                                        mode = StringOps.DefaultObjectMatchMode;

                                                    //
                                                    // NOTE: Ok, remove the specified interfaces now.
                                                    //
                                                    code = interpreter.RemoveObjectInterfaces(
                                                        mode, pattern, noCase, values, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        result = String.Empty;
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object undeclare ?options?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object undeclare ?options?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "unimport":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-bycontainer", null),
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-importmode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-importpattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-importnocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex == Index.Invalid)
                                                {
                                                    //
                                                    // NOTE: What match mode are we going to use to restrict the imported
                                                    //       namespaces, if any?
                                                    //
                                                    Variant value = null;
                                                    MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-importmode", ref value))
                                                        mode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict the imported
                                                    //       namespaces, if any?
                                                    //
                                                    string pattern = null;

                                                    if (options.IsPresent("-importpattern", ref value))
                                                        pattern = value.ToString();

                                                    //
                                                    // NOTE: For the purposes of matching namespace names, do we care
                                                    //       about case?
                                                    //
                                                    bool noCase = false;

                                                    if (options.IsPresent("-importnocase"))
                                                        noCase = true;

                                                    bool values = false;

                                                    if (options.IsPresent("-bycontainer"))
                                                        values = true;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((pattern != null) && (mode == MatchMode.None))
                                                        mode = StringOps.DefaultObjectMatchMode;

                                                    //
                                                    // NOTE: Ok, remove the specified namespaces now.
                                                    //
                                                    code = interpreter.RemoveObjectNamespaces(
                                                        mode, pattern, noCase, values, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        result = String.Empty;
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object unimport ?options?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object unimport ?options?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "untype":
                                    {
                                        if (arguments.Count >= 2)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveMatchModeValue, Index.Invalid, Index.Invalid, "-typemode", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-typepattern", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-typenocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            if (arguments.Count > 2)
                                                code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);
                                            else
                                                code = ReturnCode.Ok;

                                            if (code == ReturnCode.Ok)
                                            {
                                                if (argumentIndex == Index.Invalid)
                                                {
                                                    //
                                                    // NOTE: What match mode are we going to use to restrict, if any?
                                                    //
                                                    Variant value = null;
                                                    MatchMode mode = MatchMode.None; /* NOTE: No restriction. */

                                                    if (options.IsPresent("-typemode", ref value))
                                                        mode = (MatchMode)value.Value;

                                                    //
                                                    // NOTE: What pattern are we going to use to restrict, if any?
                                                    //
                                                    string pattern = null;

                                                    if (options.IsPresent("-typepattern", ref value))
                                                        pattern = value.ToString();

                                                    //
                                                    // NOTE: For the purposes of matching names, do we care about
                                                    //       case?
                                                    //
                                                    bool noCase = false;

                                                    if (options.IsPresent("-typenocase"))
                                                        noCase = true;

                                                    //
                                                    // NOTE: Check for a pattern without a mode (change to default,
                                                    //       which is Glob).
                                                    //
                                                    if ((pattern != null) && (mode == MatchMode.None))
                                                        mode = StringOps.DefaultObjectMatchMode;

                                                    //
                                                    // NOTE: Ok, remove the specified names now.
                                                    //
                                                    code = interpreter.RemoveObjectTypes(mode, pattern, noCase, ref result);

                                                    if (code == ReturnCode.Ok)
                                                        result = String.Empty;
                                                }
                                                else
                                                {
                                                    result = "wrong # args: should be \"object untype ?options?\"";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object untype ?options?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "verifyall":
                                    {
                                        if ((arguments.Count == 2) || (arguments.Count == 3))
                                        {
                                            bool strict = false;

                                            if (arguments.Count == 3)
                                                code = Value.GetBoolean2(
                                                    arguments[2], ValueFlags.AnyBoolean, false,
                                                    interpreter.CultureInfo, ref strict, ref result);

                                            X509VerificationFlags verificationFlags =
                                                CertificateOps.DefaultVerificationFlags;

                                            X509RevocationMode revocationMode =
                                                X509RevocationMode.Online;

                                            X509RevocationFlag revocationFlag =
                                                X509RevocationFlag.ExcludeRoot;

                                            if (code == ReturnCode.Ok)
                                            {
                                                AppDomain appDomain = interpreter.GetAppDomain();

                                                if (appDomain != null)
                                                {
                                                    Assembly[] assemblies = appDomain.GetAssemblies();

                                                    if (assemblies != null)
                                                    {
                                                        foreach (Assembly assembly in assemblies)
                                                        {
                                                            if (assembly != null)
                                                            {
                                                                if (strict || !assembly.GlobalAssemblyCache)
                                                                {
                                                                    StrongName strongName = null;

                                                                    code = AssemblyOps.GetStrongName(
                                                                        assembly, ref strongName, ref result);

                                                                    if (code != ReturnCode.Ok)
                                                                    {
                                                                        result = String.Format(
                                                                            "assembly \"{0}\": {1}",
                                                                            assembly, result);

                                                                        break;
                                                                    }

                                                                    if (!RuntimeOps.IsStrongNameVerified(
                                                                            assembly.Location, true))
                                                                    {
                                                                        result = String.Format(
                                                                            "assembly \"{0}\": not fully-signed",
                                                                            assembly);

                                                                        code = ReturnCode.Error;
                                                                        break;
                                                                    }

                                                                    X509Certificate2 certificate = null;

                                                                    code = AssemblyOps.GetCertificate2(
                                                                        assembly, true, ref certificate, ref result);

                                                                    if (code != ReturnCode.Ok)
                                                                    {
                                                                        result = String.Format(
                                                                            "assembly \"{0}\": {1}",
                                                                            assembly, result);

                                                                        break;
                                                                    }

                                                                    code = CertificateOps.VerifyChain(
                                                                        assembly, certificate, verificationFlags,
                                                                        revocationMode, revocationFlag, ref result);

                                                                    if (code != ReturnCode.Ok)
                                                                        break;

                                                                    if (!RuntimeOps.IsFileTrusted(
                                                                            assembly.Location))
                                                                    {
                                                                        result = String.Format(
                                                                            "assembly \"{0}\": is not trusted",
                                                                            assembly);

                                                                        code = ReturnCode.Error;
                                                                        break;
                                                                    }
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = "invalid assembly";
                                                                code = ReturnCode.Error;
                                                                break;
                                                            }
                                                        }

                                                        if (code == ReturnCode.Ok)
                                                            result = String.Empty;
                                                    }
                                                    else
                                                    {
                                                        result = "invalid assemblies";
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    result = "invalid application domain";
                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"object verifyall ?strict?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"object option ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}

