/*
 * GlobalState.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Reflection;

#if CAS_POLICY
using System.Security.Policy;
#endif

using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Private
{
    [ObjectId("e8491fec-2fd3-455e-92fd-cf2a84c75e8a")]
    internal static class GlobalState
    {
        #region Private Read-Only Data (Logical Constants)
        private static readonly object syncRoot = new object();

        ///////////////////////////////////////////////////////////////////////

        #region Default Package Name & Version Data
        private static readonly string DefaultPackageName = "Eagle";

        //
        // TODO: This version information should be changed if the major or
        //       minor version of the assembly changes.
        //
        private static readonly int DefaultMajorVersion = 1;
        private static readonly int DefaultMinorVersion = 0;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Entry & Executing Assembly Data
        #region Entry Assembly Data
        private static readonly Assembly entryAssembly =
            Assembly.GetEntryAssembly();

        private static readonly AssemblyName entryAssemblyName =
            (entryAssembly != null) ? entryAssembly.GetName() : null;

        private static readonly string entryAssemblyTitle =
            AttributeOps.GetAssemblyTitle(entryAssembly);

        private static readonly string entryAssemblyLocation =
            (entryAssembly != null) ? entryAssembly.Location : null;

        private static readonly Version entryAssemblyVersion =
            (entryAssemblyName != null) ? entryAssemblyName.Version : null;

        ///////////////////////////////////////////////////////////////////////

        private static readonly string entryAssemblyPath =
            AssemblyOps.GetPath(entryAssembly);
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Executing Assembly Data
        private static readonly Assembly thisAssembly =
            Assembly.GetExecutingAssembly();

#if CAS_POLICY
        private static readonly Evidence thisAssemblyEvidence =
            (thisAssembly != null) ? thisAssembly.Evidence : null;
#endif

        private static readonly AssemblyName thisAssemblyName =
            (thisAssembly != null) ? thisAssembly.GetName() : null;

        private static readonly string thisAssemblyTitle =
            AttributeOps.GetAssemblyTitle(thisAssembly);

        private static readonly string thisAssemblyLocation =
            (thisAssembly != null) ? thisAssembly.Location : null;

        private static readonly string thisAssemblySimpleName =
            (thisAssemblyName != null) ? thisAssemblyName.Name : null;

        private static readonly string thisAssemblyFullName =
            (thisAssemblyName != null) ? thisAssemblyName.FullName : null;

        private static readonly Version thisAssemblyVersion =
            (thisAssemblyName != null) ? thisAssemblyName.Version : null;

        private static readonly CultureInfo thisAssemblyCultureInfo =
            (thisAssemblyName != null) ? thisAssemblyName.CultureInfo : null;

        ///////////////////////////////////////////////////////////////////////

        private static readonly string thisAssemblyPath =
            AssemblyOps.GetPath(thisAssembly);

        private static readonly Uri thisAssemblyUri =
            AttributeOps.GetAssemblyUri(thisAssembly);

        //
        // TODO: Change this if the XSD schema URI changes.
        //
        private static readonly Uri thisAssemblyNamespaceUri =
            (thisAssemblyUri != null) ?
                new Uri(thisAssemblyUri, "2009/schema") : null;

        ///////////////////////////////////////////////////////////////////////

        #region Package Data
        private static readonly string packageName =
            GetPackageName(false);

        //
        // HACK: The package version *IS* the assembly version.
        //
        private static readonly Version packageVersion =
            thisAssemblyVersion;

        ///////////////////////////////////////////////////////////////////////

#if UNIX
        private static readonly string unixPackageName =
            GetPackageName(true);

        //
        // HACK: The Unix package version *IS* the assembly version.
        //
        private static readonly Version unixPackageVersion =
            thisAssemblyVersion;
#endif

        ///////////////////////////////////////////////////////////////////////

#if DEBUGGER
        private static readonly string debuggerName =
            String.Format("{0} {1}", packageName, Debugger.Name).Trim();
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Binary Executable Data
        private static readonly string binaryPath = PathOps.GetBinaryPath(true);
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Resource Management Data
        private static readonly string resourceBaseName =
            GetPackageName(false);
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Read-Only Primary Thread Data (Logical Constants)
        //
        // HACK: This is done dead last so we can output a useful diagnostic
        //       message showing that the global state has been initialized
        //       successfully.
        //
        private static readonly int primaryThreadId = GetCurrentThreadId();

        ///////////////////////////////////////////////////////////////////////

        private static readonly Thread primaryThread =
            SetupPrimaryThread(primaryThreadId);
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Read-Write Data
        #region Diagnostic Data
        //
        // NOTE: When this is non-zero, diagnostic messages will be written
        //       whenever the simpler overloads of the GetEnvironmentVariable
        //       and SetEnvironmentVariable methods are called.
        //
        private static bool verboseEnvironment = Build.Debug;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Object Identity Data
#if RANDOMIZE_ID
        private static Random random = new Random();
        private static long nextId = Math.Abs(random.Next());

#if !SHARED_ID_POOL
        private static long nextInterpreterId = Math.Abs(random.Next());
#endif
#else
        private static long nextId = 0;

#if !SHARED_ID_POOL
        private static long nextInterpreterId = 0;
#endif
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Threading Data
        [ThreadStatic()] /* ThreadSpecificData */
        private static int currentThreadId;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region "Active" / "All" Interpreter Tracking
        [ThreadStatic()] /* ThreadSpecificData */
        private static InterpreterStackList activeInterpreters;

        private static readonly InterpreterDictionary allInterpreters =
            new InterpreterDictionary();
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Path Override Data
        #region Environment Variable Path Overrides
        private static string libraryPath =
            GetEnvironmentVariable(EnvVars.EagleLibrary, true, false);

        private static StringList autoPathList = StringList.FromString(
            GetEnvironmentVariable(EnvVars.EagleLibPath, true, false));

        private static string tclLibraryPath =
            GetEnvironmentVariable(EnvVars.TclLibrary, true, false);

        private static StringList tclAutoPathList = StringList.FromString(
            GetEnvironmentVariable(EnvVars.TclLibPath, true, false));
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Master Base / Library Path Overrides
        //
        // NOTE: Master override for the base path.
        //
        private static string masterBasePath = null;

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: Master override for the library path.
        //
        private static string masterLibraryPath = null;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Package Path Data
        private static string assemblyPackageNamePath =
            GetAssemblyPackagePath(packageName, packageVersion);

        private static string assemblyPackageRootPath =
            GetAssemblyPackagePath(null, null);

        ///////////////////////////////////////////////////////////////////////

        private static string binaryPackageNamePath =
            GetBinaryPackagePath(packageName, packageVersion);

        private static string binaryPackageRootPath =
            GetBinaryPackagePath(null, null);

        ///////////////////////////////////////////////////////////////////////

        private static string packagePeerPath = GetPackagePath(
            thisAssembly, null, null, false);

        private static string packageRootPath = GetPackagePath(
            thisAssembly, null, null, true);

        ///////////////////////////////////////////////////////////////////////

        private static string packageNamePath = GetPackagePath(
            thisAssembly, packageName, packageVersion, false);

        private static string packageNameRootPath = GetPackagePath(
            thisAssembly, packageName, packageVersion, true);

        ///////////////////////////////////////////////////////////////////////

#if UNIX
        private static string unixPackageNameLocalPath = GetUnixPackagePath(
            unixPackageName, unixPackageVersion, true);

        private static string unixPackageNamePath = GetUnixPackagePath(
            unixPackageName, unixPackageVersion, false);
#endif

        ///////////////////////////////////////////////////////////////////////

#if NATIVE && TCL
        private static string tclPackageNamePath =
            GetPackagePath(thisAssembly, TclVars.PackageName, null, false);

        private static string tclPackageNameRootPath =
            GetPackagePath(thisAssembly, TclVars.PackageName, null, true);
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Master Auto-Path List
        //
        // NOTE: Master auto-path list.  This list is automatically initialized
        //       [once] when necessary; however, it may then be overridden to
        //       influence the behavior of any interpreters created after it
        //       has been changed.
        //
        private static StringList masterAutoPathList = null;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Object Identity Methods
        public static long NextId() /* THREAD-SAFE */
        {
            long result;

            //
            // NOTE: This is our cheap unique Id generator for
            //       the various script visible identifiers
            //       (such as channel names, etc).  This value
            //       is not per-interpreter; therefore, use with
            //       caution.
            //
            result = Interlocked.Increment(ref nextId);

            if (result < 0)
                DebugOps.Complain(ReturnCode.Error,
                    "next global identifier is negative");

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static long NextId(
            Interpreter interpreter
            )
        {
            return (interpreter != null) ?
                interpreter.NextId() : NextId();
        }

        ///////////////////////////////////////////////////////////////////////

        public static long NextInterpreterId() /* THREAD-SAFE */
        {
            long result;

#if SHARED_ID_POOL
            result = NextId();
#else
            result = Interlocked.Increment(ref nextInterpreterId);
#endif

            if (result < 0)
                DebugOps.Complain(ReturnCode.Error,
                    "next interpreter identifier is negative");

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Threading Global Variable Access Methods
        public static int GetCurrentThreadId()
        {
            int threadId = currentThreadId;

            if (threadId == 0)
                currentThreadId = threadId = AppDomain.GetCurrentThreadId();

            return threadId;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Thread SetupPrimaryThread(
            int threadId
            )
        {
            Thread thread = Thread.CurrentThread;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: SetupPrimaryThread: library initialized on managed " +
                "thread with [{1}] and with next object Id {2} and with " +
                "next interpreter Id {3}.{4}", threadId,
                FormatOps.DisplayThread(thread), nextId, nextInterpreterId,
                Environment.NewLine), typeof(GlobalState).Name);

            return thread;
        }

        ///////////////////////////////////////////////////////////////////////

        public static int GetPrimaryThreadId()
        {
            return primaryThreadId;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Thread GetPrimaryThread()
        {
            return primaryThread;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region "Active" / "All" Interpreter Tracking Methods
        #region "All" Interpreter Tracking Methods
        public static ReturnCode GetInterpreter( /* NOTE: GetAnyInterpreter */
            bool validate,
            ref Interpreter interpreter,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (allInterpreters != null)
                {
                    //
                    // NOTE: Grab the first available [and valid?] interpreter.
                    //
                    foreach (KeyValuePair<string, Interpreter> pair
                            in allInterpreters)
                    {
                        interpreter = pair.Value;

                        if (!validate || (interpreter != null))
                            return ReturnCode.Ok;
                    }

                    error = String.Format(
                        "no {0}interpreter found",
                        validate ? "valid " : String.Empty);
                }
                else
                {
                    error = "no interpreters available";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetInterpreter(
            string name,
            bool validate,
            ref Interpreter interpreter,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (allInterpreters != null)
                {
                    //
                    // NOTE: *WARNING* Empty interpreter names are
                    //       technically allowed, please do not
                    //       change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (allInterpreters.ContainsKey(name))
                        {
                            interpreter = allInterpreters[name];

                            if (!validate || (interpreter != null))
                                return ReturnCode.Ok;
                            else
                                error = String.Format(
                                    "invalid interpreter name \"{0}\"",
                                    FormatOps.DisplayName(name));
                        }
                        else
                        {
                            error = String.Format(
                                "interpreter \"{0}\" not found",
                                FormatOps.DisplayName(name));
                        }
                    }
                    else
                    {
                        error = "invalid interpreter name";
                    }
                }
                else
                {
                    error = "no interpreters available";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static InterpreterDictionary GetInterpreters()
        {
            lock (syncRoot)
            {
                if (allInterpreters != null)
                    return (InterpreterDictionary)allInterpreters.Clone();
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool AddInterpreter(
            Interpreter interpreter
            )
        {
            lock (syncRoot)
            {
                if (allInterpreters != null)
                {
                    allInterpreters.Add(interpreter);

                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool RemoveInterpreter(
            Interpreter interpreter
            )
        {
            lock (syncRoot)
            {
                if (allInterpreters != null)
                    return allInterpreters.Remove(interpreter);
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region "Active" Interpreter Tracking Methods
        public static IAnyPair<Interpreter, IClientData> GetActiveInterpreter()
        {
            return GetActiveInterpreter(null);
        }

        ///////////////////////////////////////////////////////////////////////

        public static IAnyPair<Interpreter, IClientData> GetActiveInterpreter(
            Type type
            )
        {
            if ((activeInterpreters != null) && !activeInterpreters.IsEmpty)
            {
                if (type != null)
                {
                    for (int index = 0;
                            index < activeInterpreters.Count;
                            index++)
                    {
                        IAnyPair<Interpreter, IClientData> pair =
                            activeInterpreters.Peek(index);

                        if (pair == null)
                            continue;

                        IClientData clientData = pair.Y;

                        if (clientData == null)
                            continue;

                        if (clientData.GetType() == type)
                            return pair;
                    }
                }
                else
                {
                    return activeInterpreters.Peek();
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static InterpreterStackList GetActiveInterpreters()
        {
            if (activeInterpreters != null)
                return (InterpreterStackList)activeInterpreters.Clone();

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void PushActiveInterpreter(
            Interpreter interpreter
            )
        {
            PushActiveInterpreter(interpreter, null);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void PushActiveInterpreter(
            Interpreter interpreter,
            IClientData clientData
            )
        {
            int pushed = 0;

            PushActiveInterpreter(interpreter, clientData, ref pushed);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void PushActiveInterpreter(
            Interpreter interpreter,
            IClientData clientData,
            ref int pushed
            )
        {
            if (activeInterpreters == null)
                activeInterpreters = new InterpreterStackList();

            activeInterpreters.Push(new AnyPair<Interpreter, IClientData>(
                interpreter, clientData));

            Interlocked.Increment(ref pushed);

#if NOTIFY && NOTIFY_GLOBAL && NOTIFY_ACTIVE
            if ((interpreter != null) && interpreter.GlobalNotify)
            {
                /* IGNORED */
                Interpreter.CheckNotifications(null, false,
                    NotifyType.Interpreter, NotifyFlags.Pushed,
                    null, interpreter, clientData, null, null);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static IAnyPair<Interpreter, IClientData> PopActiveInterpreter()
        {
            int pushed = 1; // required, or no pop.

            return PopActiveInterpreter(ref pushed);
        }

        ///////////////////////////////////////////////////////////////////////

        public static IAnyPair<Interpreter, IClientData> PopActiveInterpreter(
            ref int pushed
            )
        {
            IAnyPair<Interpreter, IClientData> pair = null;

            if (pushed > 0)
            {
                if ((activeInterpreters != null) &&
                    !activeInterpreters.IsEmpty)
                {
                    pair = activeInterpreters.Pop();
                    Interlocked.Decrement(ref pushed);
                }
            }

#if NOTIFY && NOTIFY_GLOBAL && NOTIFY_ACTIVE
            if (pair != null)
            {
                Interpreter interpreter = pair.X;

                if ((interpreter != null) && interpreter.GlobalNotify)
                {
                    IClientData clientData = pair.Y;

                    /* IGNORED */
                    Interpreter.CheckNotifications(null, false,
                        NotifyType.Interpreter, NotifyFlags.Popped,
                        null, interpreter, clientData, null, null);
                }
            }
#endif

            return pair;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Environment Variable Management Methods
        public static string GetEnvironmentVariable(
            string variable,
            bool configuration,
            bool expand
            )
        {
            return GetEnvironmentVariable(
                variable, configuration, expand, verboseEnvironment);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetEnvironmentVariable(
            string variable,
            bool configuration,
            bool expand,
            bool verbose
            )
        {
            string value = null;

            //
            // NOTE: If the variable name is null or empty, return the default
            //       value (null) instead of potentially throwing an exception
            //       later.
            //
            if (String.IsNullOrEmpty(variable))
                return value;

            //
            // NOTE: If this is a "configuration" setting, try for the variable
            //       name prefixed by our package name first.
            //
            if (configuration)
                value = CommonOps.GetEnvironmentVariable(String.Format(
                    "{0}_{1}", GetPackageName(false), variable));

            //
            // NOTE: Failing to find that, just try for the variable name.
            //
            if (value == null)
                value = CommonOps.GetEnvironmentVariable(variable);

            //
            // NOTE: If necessary, expand any contained environment variables.
            //
            if (expand && !String.IsNullOrEmpty(value))
                value = CommonOps.ExpandEnvironmentVariables(value);

            //
            // NOTE: Output a diagnostic message about the environment variable
            //       request.
            //
            if (verbose)
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: GetEnvironmentVariable: variable = {1}, " +
                    "configuration = {2}, expand = {3}, value = {4}{5}",
                    GetCurrentThreadId(), FormatOps.WrapOrNull(variable),
                    configuration, expand, FormatOps.WrapOrNull(value),
                    Environment.NewLine), typeof(GlobalState).Name);

            return value;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void SetEnvironmentVariable(
            string variable,
            string value
            )
        {
            SetEnvironmentVariable(variable, value, verboseEnvironment);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void SetEnvironmentVariable(
            string variable,
            string value,
            bool verbose
            )
        {
            //
            // NOTE: Attempt to set the requested environment variable.
            //
            CommonOps.SetEnvironmentVariable(variable, value);

            //
            // NOTE: Output a diagnostic message about the environment variable
            //       request.
            //
            if (verbose)
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: SetEnvironmentVariable: variable = {1}, value = {2}{3}",
                    GetCurrentThreadId(), FormatOps.WrapOrNull(variable),
                    FormatOps.WrapOrNull(value), Environment.NewLine),
                    typeof(GlobalState).Name);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Assembly Global Variable Access Methods
        #region Entry Assembly Methods
        public static Assembly GetEntryAssembly()
        {
            return entryAssembly;
        }

        ///////////////////////////////////////////////////////////////////////

        public static AssemblyName GetEntryAssemblyName()
        {
            return entryAssemblyName;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetEntryAssemblyTitle()
        {
            return entryAssemblyTitle;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetEntryAssemblyLocation()
        {
            return entryAssemblyLocation;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Version GetEntryAssemblyVersion()
        {
            return entryAssemblyVersion;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetEntryAssemblyPath()
        {
            return entryAssemblyPath;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Executing Assembly Methods
        public static Assembly GetAssembly()
        {
            return thisAssembly;
        }

        ///////////////////////////////////////////////////////////////////////

#if CAS_POLICY
        public static Evidence GetAssemblyEvidence()
        {
            return thisAssemblyEvidence;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static AssemblyName GetAssemblyName()
        {
            return thisAssemblyName;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblyTitle()
        {
            return thisAssemblyTitle;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblyLocation()
        {
            return thisAssemblyLocation;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblySimpleName()
        {
            return thisAssemblySimpleName;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblyFullName()
        {
            return thisAssemblyFullName;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Version GetAssemblyVersion()
        {
            return thisAssemblyVersion;
        }

        ///////////////////////////////////////////////////////////////////////

        public static CultureInfo GetAssemblyCultureInfo()
        {
            return thisAssemblyCultureInfo;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblyPath()
        {
            return thisAssemblyPath;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Uri GetAssemblyUri()
        {
            return thisAssemblyUri;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Uri GetAssemblyNamespaceUri()
        {
            return thisAssemblyNamespaceUri;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Binary Executable Methods
        public static string GetBinaryPath()
        {
            return binaryPath;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Resource Management Methods
        public static string GetResourceBaseName()
        {
            return resourceBaseName;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Package Global Variable Access Methods
#if DEBUGGER
        public static string GetDebuggerName()
        {
            return debuggerName;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static string GetPackageName( /* CANNOT RETURN NULL */
            bool noCase
            )
        {
            return GetPackageName(null, null, noCase);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetPackageName( /* CANNOT RETURN NULL */
            string prefix,
            string suffix,
            bool noCase
            )
        {
            //
            // NOTE: We do not want to return the assembly name here because
            //       that is not guaranteed to be the same as what we consider
            //       the "package name".
            //
            string result = GetAssemblyTitle();

            if (String.IsNullOrEmpty(result))
                result = DefaultPackageName;

            if (noCase)
                result = result.ToLowerInvariant();

            if (!String.IsNullOrEmpty(prefix) || !String.IsNullOrEmpty(suffix))
                result = String.Format("{0}{1}{2}", prefix, result, suffix);

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetPackageNameRootPath()
        {
            lock (syncRoot)
            {
                return packageNameRootPath;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static Version GetPackageVersion() /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                //
                // NOTE: Package versions do not typically include the build
                //       and revision numbers; therefore, be sure they are
                //       omitted in our return value.
                //
                return (packageVersion != null) ?
                    new Version(packageVersion.Major,
                        packageVersion.Minor) :
                    new Version(DefaultMajorVersion, DefaultMinorVersion);
            }
        }

        ///////////////////////////////////////////////////////////////////////

#if NATIVE && TCL
        public static string GetTclPackageNamePath()
        {
            lock (syncRoot)
            {
                return tclPackageNamePath;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetTclPackageNameRootPath()
        {
            lock (syncRoot)
            {
                return tclPackageNameRootPath;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Base Path Global Variable Management Methods
        public static string GetBasePath()
        {
            return GetBasePath(thisAssembly);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetBasePath(
            Assembly assembly /* OPTIONAL: May be null. */
            ) /* THREAD-SAFE */
        {
            string result = null;

            lock (syncRoot)
            {
                if (!String.IsNullOrEmpty(masterBasePath))
                {
                    //
                    // NOTE: Allow manual override of the base path.
                    //
                    result = masterBasePath;
                }
                else if ((assembly != null) && assembly.GlobalAssemblyCache)
                {
                    //
                    // NOTE: The assembly has been GAC'd.  We need
                    //       to use the registry to find where we
                    //       were actually installed to.
                    //
                    result = SetupOps.GetPath(packageVersion);

                    //
                    // NOTE: If we failed to get the path from the
                    //       setup registry hive (perhaps setup was
                    //       not run?) then we resort to using the
                    //       current application probing path.
                    //
                    if (String.IsNullOrEmpty(result) &&
                        !String.IsNullOrEmpty(binaryPath))
                    {
                        result = PathOps.GetBasePath(assembly, binaryPath);
                    }
                }
                else
                {
                    //
                    // NOTE: Return the base directory that this
                    //       assembly was loaded from (i.e. without
                    //       the "bin").
                    //
                    if (!String.IsNullOrEmpty(binaryPath))
                        result = PathOps.GetBasePath(assembly, binaryPath);
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void SetBasePath(
            string basePath,
            bool refresh
            ) /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                masterBasePath = basePath;

                //
                // BUGFIX: Be sure to propagate the changes down to where they
                //         are actually useful.
                //
                if (refresh)
                    RefreshBasePath();
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void RefreshBasePath()
        {
            RefreshLibraryPath();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Library Path Global Variable Management Methods
        public static string GetLibraryPath(
            Assembly assembly, /* OPTIONAL: May be null. */
            bool root
            ) /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                if (!String.IsNullOrEmpty(masterLibraryPath))
                {
                    //
                    // NOTE: Allow manual override of the library path.
                    //
                    return masterLibraryPath;
                }
                else if (root)
                {
                    //
                    // NOTE: We want the root library directory.  This path
                    //       allows us to run from the build directory (e.g.
                    //       "bin\Debug\bin") and still refer to directories
                    //       that are not in the build directory (i.e. they
                    //       are a peer of the outer "bin" directory).
                    //
                    return Path.GetFullPath(PathOps.CombinePath(
                        false, GetBasePath(assembly), TclVars.LibPath));
                }
                else
                {
                    //
                    // NOTE: We want the non-root (or peer) library directory.
                    //       When running in a non-build environment, this will
                    //       typically be the same as the root library directory.
                    //
                    // BUGBUG: This basically assumes that the directory
                    //         containing the application binary is parallel to
                    //         the Eagle library directory.  This will not work
                    //         if Eagle is running from "/usr/local/bin/Eagle"
                    //         and the library directory is "/usr/local/lib/Eagle/"
                    //         (OpenBSD).  In order for that layout to work, we
                    //         would have to go up one more level.
                    //
                    return Path.GetFullPath(PathOps.CombinePath(
                        false, Path.GetDirectoryName(binaryPath),
                        TclVars.LibPath));
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void SetLibraryPath(
            string libraryPath,
            bool refresh
            ) /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                masterLibraryPath = libraryPath;

                //
                // BUGFIX: Be sure to propagate the changes down to where they
                //         are actually useful.
                //
                if (refresh)
                    RefreshLibraryPath();
            }
        }

        ///////////////////////////////////////////////////////////////////////

#if UNIX
        public static string GetUnixLibraryPath(
            bool local
            ) /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                if (!String.IsNullOrEmpty(masterLibraryPath))
                {
                    //
                    // NOTE: Allow manual override of the library path.
                    //
                    return masterLibraryPath;
                }
                else if (local)
                {
                    //
                    // NOTE: We want the directory where local libraries
                    //       are installed.
                    //
                    return TclVars.UserLocalLibPath;
                }
                else
                {
                    //
                    // NOTE: We want the directory where libraries are
                    //       installed.
                    //
                    return TclVars.UserLibPath;
                }
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static void RefreshLibraryPath() /* THREAD-SAFE */
        {
            lock (syncRoot)
            {
                packagePeerPath = GetPackagePath(
                    thisAssembly, null, null, false);

                packageRootPath = GetPackagePath(
                    thisAssembly, null, null, true);

                ///////////////////////////////////////////////////////////////

                packageNamePath = GetPackagePath(
                    thisAssembly, packageName, packageVersion, false);

                packageNameRootPath = GetPackagePath(
                    thisAssembly, packageName, packageVersion, true);

                ///////////////////////////////////////////////////////////////

#if UNIX
                unixPackageNameLocalPath = GetUnixPackagePath(
                    unixPackageName, unixPackageVersion, true);

                unixPackageNamePath = GetUnixPackagePath(
                    unixPackageName, unixPackageVersion, false);
#endif

                ///////////////////////////////////////////////////////////////

#if NATIVE && TCL
                tclPackageNamePath = GetPackagePath(
                    thisAssembly, TclVars.PackageName, null, false);

                tclPackageNameRootPath = GetPackagePath(
                    thisAssembly, TclVars.PackageName, null, true);
#endif

                ///////////////////////////////////////////////////////////////

                //
                // BUGFIX: Reset the master auto-path so that it will be
                //         initialized again [using our new paths] on the
                //         next call to the GetAutoPathList method.
                //
                masterAutoPathList = null;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Package Path Global Variable Management Methods
        public static string GetPackagePath(
            Assembly assembly, /* OPTIONAL: May be null. */
            string name,       /* OPTIONAL: May be null. */
            Version version,   /* OPTIONAL: May be null. */
            bool root
            )
        {
            string result = GetLibraryPath(assembly, root);

            if (!String.IsNullOrEmpty(name))
            {
                result = PathOps.CombinePath(false, result, name);

                if (version != null)
                    result += FormatOps.MajorMinor(version);
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

#if UNIX
        public static string GetUnixPackagePath(
            string name,
            Version version,
            bool local
            )
        {
            string result = GetUnixLibraryPath(local);

            if (!String.IsNullOrEmpty(name))
            {
                result = PathOps.CombinePath(true, result, name);

                if (version != null)
                    result += FormatOps.MajorMinor(version);
            }

            return result;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static string GetAssemblyPackagePath(
            string name,
            Version version
            )
        {
            string result = null;

            if (!String.IsNullOrEmpty(thisAssemblyPath))
            {
                string basePath = PathOps.GetBasePath(
                    thisAssembly, thisAssemblyPath);

                if (!String.IsNullOrEmpty(basePath))
                {
                    result = PathOps.CombinePath(
                        false, basePath, TclVars.LibPath);

                    if (!String.IsNullOrEmpty(name))
                    {
                        result = PathOps.CombinePath(false, result, name);

                        if (version != null)
                            result += FormatOps.MajorMinor(version);
                    }
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetBinaryPackagePath(
            string name,
            Version version
            )
        {
            string result = null;

            if (!String.IsNullOrEmpty(binaryPath))
            {
                string basePath = PathOps.GetBasePath(
                    thisAssembly, binaryPath);

                if (!String.IsNullOrEmpty(basePath))
                {
                    result = PathOps.CombinePath(
                        false, basePath, TclVars.LibPath);

                    if (!String.IsNullOrEmpty(name))
                    {
                        result = PathOps.CombinePath(false, result, name);

                        if (version != null)
                            result += FormatOps.MajorMinor(version);
                    }
                }
            }

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Auto-Path Global Variable Management Methods
        public static void RefreshAutoPathList(
            bool verbose
            )
        {
            lock (syncRoot)
            {
                //
                // NOTE: Re-query all our environment variables now.
                //
                libraryPath = GetEnvironmentVariable(EnvVars.EagleLibrary,
                    true, false, verbose);

                autoPathList = StringList.FromString(GetEnvironmentVariable(
                    EnvVars.EagleLibPath, true, false, verbose));

                ///////////////////////////////////////////////////////////////

                tclLibraryPath = GetEnvironmentVariable(EnvVars.TclLibrary,
                    true, false, verbose);

                tclAutoPathList = StringList.FromString(GetEnvironmentVariable(
                    EnvVars.TclLibPath, true, false, verbose));

                ///////////////////////////////////////////////////////////////

                //
                // BUGFIX: Reset the master auto-path so that it will be
                //         initialized again [using our new paths] on the
                //         next call to the GetAutoPathList method.
                //
                masterAutoPathList = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList GetAutoPathList( /* CANNOT RETURN NULL */
            Interpreter interpreter,
            bool refresh
            )
        {
            return GetAutoPathList(interpreter, (interpreter != null) ?
                interpreter.CreateFlags : CreateFlags.None, refresh);
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList GetAutoPathList( /* CANNOT RETURN NULL */
            Interpreter interpreter,
            CreateFlags createFlags,
            bool refresh
            ) /* THREAD-SAFE */
        {
            bool showAutoPath = FlagOps.HasFlags(
                createFlags, CreateFlags.ShowAutoPath, true);

            if (showAutoPath)
            {
                RuntimeOps.DebugWrite(interpreter, String.Format(
                    "GetAutoPathList: entered, interpreter = {0}, " +
                    "createFlags = {1}, refresh = {2}",
                    FormatOps.WrapOrNull(interpreter),
                    FormatOps.WrapOrNull(createFlags), refresh), true);
            }

            bool strict = FlagOps.HasFlags(
                createFlags, CreateFlags.StrictAutoPath, true);

            PathDictionary<object> paths = new PathDictionary<object>();

            if (interpreter != null)
            {
                string libraryPath = interpreter.LibraryPath; /* throw */

                if (!String.IsNullOrEmpty(libraryPath) &&
                    !paths.ContainsKey(libraryPath) &&
                    (!strict || Directory.Exists(libraryPath)))
                {
                    paths.Add(libraryPath, null);

                    if (showAutoPath)
                    {
                        RuntimeOps.DebugWrite(interpreter, String.Format(
                            "GetAutoPathList: Added interpreter library path: {0}",
                            FormatOps.WrapOrNull(libraryPath)), true);
                    }
                }

                StringList autoPathList = interpreter.AutoPathList; /* throw */

                if (autoPathList != null)
                {
                    foreach (string path in autoPathList)
                    {
                        if (!String.IsNullOrEmpty(path) &&
                            !paths.ContainsKey(path) &&
                            (!strict || Directory.Exists(path)))
                        {
                            paths.Add(path, null);

                            if (showAutoPath)
                            {
                                RuntimeOps.DebugWrite(interpreter, String.Format(
                                    "GetAutoPathList: Added interpreter package path: {0}",
                                    FormatOps.WrapOrNull(path)), true);
                            }
                        }
                    }
                }
            }

            lock (syncRoot)
            {
                if (refresh || (masterAutoPathList == null))
                {
                    if (showAutoPath)
                    {
                        RuntimeOps.DebugWrite(interpreter, String.Format(
                            "GetAutoPathList: Master path list {0}",
                            (masterAutoPathList != null) ? "was initialized" :
                                "was not initialized"), true);
                    }

                    //
                    // TODO: Add any extra "default" package index search paths
                    //       here.
                    //
                    PathDictionary<object> masterPaths = new PathDictionary<object>();

                    if (!String.IsNullOrEmpty(libraryPath) &&
                        !masterPaths.ContainsKey(libraryPath) &&
                        (!strict || Directory.Exists(libraryPath)))
                    {
                        masterPaths.Add(libraryPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added library path: {0}",
                                FormatOps.WrapOrNull(libraryPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(tclLibraryPath) &&
                        !masterPaths.ContainsKey(tclLibraryPath) &&
                        (!strict || Directory.Exists(tclLibraryPath)))
                    {
                        masterPaths.Add(tclLibraryPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added Tcl library path: {0}",
                                FormatOps.WrapOrNull(tclLibraryPath)), true);
                        }
                    }

                    if (autoPathList != null)
                    {
                        foreach (string path in autoPathList)
                        {
                            if (!String.IsNullOrEmpty(path) &&
                                !masterPaths.ContainsKey(path) &&
                                (!strict || Directory.Exists(path)))
                            {
                                masterPaths.Add(path, null);

                                if (showAutoPath)
                                {
                                    RuntimeOps.DebugWrite(interpreter, String.Format(
                                        "GetAutoPathList: Added package path: {0}",
                                        FormatOps.WrapOrNull(path)), true);
                                }
                            }
                        }
                    }

                    if (tclAutoPathList != null)
                    {
                        foreach (string path in tclAutoPathList)
                        {
                            if (!String.IsNullOrEmpty(path) &&
                                !masterPaths.ContainsKey(path) &&
                                (!strict || Directory.Exists(path)))
                            {
                                masterPaths.Add(path, null);

                                if (showAutoPath)
                                {
                                    RuntimeOps.DebugWrite(interpreter, String.Format(
                                        "GetAutoPathList: Added Tcl package path: {0}",
                                        FormatOps.WrapOrNull(path)), true);
                                }
                            }
                        }
                    }

#if UNIX
                    if (!String.IsNullOrEmpty(unixPackageNameLocalPath) &&
                        !masterPaths.ContainsKey(unixPackageNameLocalPath) &&
                        (!strict || Directory.Exists(unixPackageNameLocalPath)))
                    {
                        masterPaths.Add(unixPackageNameLocalPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added Unix package name local path: {0}",
                                FormatOps.WrapOrNull(unixPackageNameLocalPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(unixPackageNamePath) &&
                        !masterPaths.ContainsKey(unixPackageNamePath) &&
                        (!strict || Directory.Exists(unixPackageNamePath)))
                    {
                        masterPaths.Add(unixPackageNamePath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added Unix package name path: {0}",
                                FormatOps.WrapOrNull(unixPackageNamePath)), true);
                        }
                    }
#endif

                    if (!String.IsNullOrEmpty(packageNamePath) &&
                        !masterPaths.ContainsKey(packageNamePath) &&
                        (!strict || Directory.Exists(packageNamePath)))
                    {
                        masterPaths.Add(packageNamePath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added package name path: {0}",
                                FormatOps.WrapOrNull(packageNamePath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(packageNameRootPath) &&
                        !masterPaths.ContainsKey(packageNameRootPath) &&
                        (!strict || Directory.Exists(packageNameRootPath)))
                    {
                        masterPaths.Add(packageNameRootPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added package name root path: {0}",
                                FormatOps.WrapOrNull(packageNameRootPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(packagePeerPath) &&
                        !masterPaths.ContainsKey(packagePeerPath) &&
                        (!strict || Directory.Exists(packagePeerPath)))
                    {
                        masterPaths.Add(packagePeerPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added package peer path: {0}",
                                FormatOps.WrapOrNull(packagePeerPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(packageRootPath) &&
                        !masterPaths.ContainsKey(packageRootPath) &&
                        (!strict || Directory.Exists(packageRootPath)))
                    {
                        masterPaths.Add(packageRootPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added package root path: {0}",
                                FormatOps.WrapOrNull(packageRootPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(binaryPackageNamePath) &&
                        !masterPaths.ContainsKey(binaryPackageNamePath) &&
                        (!strict || Directory.Exists(binaryPackageNamePath)))
                    {
                        masterPaths.Add(binaryPackageNamePath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added binary package name path: {0}",
                                FormatOps.WrapOrNull(binaryPackageNamePath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(binaryPackageRootPath) &&
                        !masterPaths.ContainsKey(binaryPackageRootPath) &&
                        (!strict || Directory.Exists(binaryPackageRootPath)))
                    {
                        masterPaths.Add(binaryPackageRootPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added binary package root path: {0}",
                                FormatOps.WrapOrNull(binaryPackageRootPath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(assemblyPackageNamePath) &&
                        !masterPaths.ContainsKey(assemblyPackageNamePath) &&
                        (!strict || Directory.Exists(assemblyPackageNamePath)))
                    {
                        masterPaths.Add(assemblyPackageNamePath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added assembly package name path: {0}",
                                FormatOps.WrapOrNull(assemblyPackageNamePath)), true);
                        }
                    }

                    if (!String.IsNullOrEmpty(assemblyPackageRootPath) &&
                        !masterPaths.ContainsKey(assemblyPackageRootPath) &&
                        (!strict || Directory.Exists(assemblyPackageRootPath)))
                    {
                        masterPaths.Add(assemblyPackageRootPath, null);

                        if (showAutoPath)
                        {
                            RuntimeOps.DebugWrite(interpreter, String.Format(
                                "GetAutoPathList: Added assembly package root path: {0}",
                                FormatOps.WrapOrNull(assemblyPackageRootPath)), true);
                        }
                    }

                    masterAutoPathList = new StringList(masterPaths.Keys);

                    if (showAutoPath)
                    {
                        RuntimeOps.DebugWrite(interpreter, String.Format(
                            "GetAutoPathList: Master path list initialized to: {0}",
                            FormatOps.WrapOrNull(masterAutoPathList)), true);
                    }
                }
                else
                {
                    if (showAutoPath)
                    {
                        RuntimeOps.DebugWrite(interpreter,
                            "GetAutoPathList: Master path list already initialized",
                            true);
                    }
                }

                if (showAutoPath)
                {
                    RuntimeOps.DebugWrite(interpreter, String.Format(
                        "GetAutoPathList: exited, interpreter = {0}, " +
                        "createFlags = {1}, refresh = {2}",
                        FormatOps.WrapOrNull(interpreter),
                        FormatOps.WrapOrNull(createFlags), refresh), true);
                }

                //
                // NOTE: Add the master list after the per-interpreter one.
                //
                paths.Add(masterAutoPathList, true);
            }

            return new StringList(paths.Keys);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Path Debugging Support Methods
        public static void GetPaths(
            Interpreter interpreter,
            ref StringPairList list
            )
        {
            if (list == null)
                list = new StringPairList();

            ///////////////////////////////////////////////////////////////////
            //
            // PHASE 0: Interpreter path variables.
            //
            ///////////////////////////////////////////////////////////////////

            if (interpreter != null)
            {
                list.Add("interpreter library path", interpreter.LibraryPath);

                list.Add("interpreter auto_path list",
                    (interpreter.AutoPathList != null) ?
                        interpreter.AutoPathList.ToString() : null);
            }

            ///////////////////////////////////////////////////////////////////
            //
            // PHASE 1: Basic path variables.
            //
            ///////////////////////////////////////////////////////////////////

            list.Add("binary path", binaryPath);
            list.Add("assembly path", thisAssemblyPath);

            list.Add("master base path", masterBasePath);
            list.Add("master library path", masterLibraryPath);

            list.Add("library path", libraryPath);
            list.Add("Tcl library path", tclLibraryPath);

            list.Add("package name path", packageNamePath);
            list.Add("package name root path", packageNameRootPath);

#if NATIVE && TCL
            list.Add("Tcl package name path", tclPackageNamePath);
            list.Add("Tcl package name root path", tclPackageNameRootPath);
#endif

#if UNIX
            list.Add("Unix package name local path", unixPackageNameLocalPath);
            list.Add("Unix package name path", unixPackageNamePath);
#endif

            list.Add("package peer path", packagePeerPath);
            list.Add("package root path", packageRootPath);

            list.Add("binary package name path", binaryPackageNamePath);
            list.Add("binary package root path", binaryPackageRootPath);

            list.Add("assembly package name path", assemblyPackageNamePath);
            list.Add("assembly package root path", assemblyPackageRootPath);

            ///////////////////////////////////////////////////////////////////
            //
            // PHASE 2: Path lists.
            //
            ///////////////////////////////////////////////////////////////////

            list.Add("auto_path list",
                (autoPathList != null) ? autoPathList.ToString() : null);

            list.Add("Tcl auto_path list",
                (tclAutoPathList != null) ? tclAutoPathList.ToString() : null);

            list.Add("master auto_path list",
                (masterAutoPathList != null) ? masterAutoPathList.ToString() : null);

            ///////////////////////////////////////////////////////////////////
            //
            // PHASE 3: Path resolution routines.
            //
            ///////////////////////////////////////////////////////////////////

            try
            {
                list.Add("Path.GetDirectoryName(binaryPath)",
                    Path.GetDirectoryName(binaryPath));
            }
            catch (Exception e)
            {
                list.Add("Path.GetDirectoryName(binaryPath)",
                    e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                string directory = Path.GetDirectoryName(binaryPath);

                list.Add("Path.GetFullPath(Path.GetDirectoryName(binaryPath))",
                    (directory != null) ? Path.GetFullPath(directory) : null);
            }
            catch (Exception e)
            {
                list.Add("Path.GetFullPath(Path.GetDirectoryName(binaryPath))",
                    e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                list.Add("GlobalState.GetBasePath(thisAssembly)",
                    GetBasePath(thisAssembly));
            }
            catch (Exception e)
            {
                list.Add("GlobalState.GetBasePath(thisAssembly)",
                    e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                list.Add("PathOps.GetBasePath(thisAssembly, thisAssemblyPath)",
                    PathOps.GetBasePath(thisAssembly, thisAssemblyPath));
            }
            catch (Exception e)
            {
                list.Add("PathOps.GetBasePath(thisAssembly, thisAssemblyPath)",
                    e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                string basePath = PathOps.GetBasePath(
                    thisAssembly, thisAssemblyPath);

                list.Add("Path.GetFullPath(PathOps.GetBasePath(" +
                    "thisAssembly, thisAssemblyPath))", (basePath != null) ?
                    Path.GetFullPath(basePath) : null);
            }
            catch (Exception e)
            {
                list.Add("Path.GetFullPath(PathOps.GetBasePath(" +
                    "thisAssembly, thisAssemblyPath))", e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                list.Add("PathOps.GetBasePath(thisAssembly, binaryPath)",
                    PathOps.GetBasePath(thisAssembly, binaryPath));
            }
            catch (Exception e)
            {
                list.Add("PathOps.GetBasePath(thisAssembly, binaryPath)",
                    e.ToString());
            }

            ///////////////////////////////////////////////////////////////////

            try
            {
                string basePath = PathOps.GetBasePath(
                    thisAssembly, binaryPath);

                list.Add("Path.GetFullPath(PathOps.GetBasePath(" +
                    "thisAssembly, binaryPath))", (basePath != null) ?
                    Path.GetFullPath(basePath) : null);
            }
            catch (Exception e)
            {
                list.Add("Path.GetFullPath(PathOps.GetBasePath(" +
                    "thisAssembly, binaryPath))", e.ToString());
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void DisplayPaths(
            Interpreter interpreter
            )
        {
            StringPairList list = null;

            GetPaths(interpreter, ref list);

            if (list != null)
            {
                foreach (IPair<string> element in list)
                {
                    if (element != null)
                    {
                        /* EXEMPT */
                        DebugOps.Write(interpreter, String.Format(
                            "DisplayPaths: {0} = {1}", element.X,
                            FormatOps.WrapOrNull(element.Y)), true);
                    }
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Embedding Support Methods
        public static bool DetectLibraryPath( /* EXTERNAL USE ONLY */
            Assembly assembly
            )
        {
            string name = GetPackageName(false);

            if (!String.IsNullOrEmpty(name))
            {
                Version version = GetPackageVersion();

                if (version != null)
                {
                    string path = GetPackagePath(assembly, name, version, true);
                    string fileName = FileName.Initialization;

                    if (String.IsNullOrEmpty(path) || !Directory.Exists(path) ||
                        !File.Exists(PathOps.CombinePath(false, path, fileName)))
                    {
                        path = CommonOps.GetEnvironmentVariable(EnvVars.Eagle);

                        if (!String.IsNullOrEmpty(path) && Directory.Exists(path))
                        {
                            path = PathOps.CombinePath(false, path,
                                FormatOps.PackageDirectory(name, version, true));

                            if (!Directory.Exists(path) ||
                                !File.Exists(PathOps.CombinePath(false, path, fileName)))
                            {
                                path = null;
                            }
                        }
                        else
                        {
                            path = null;
                        }
                    }
                    else
                    {
                        path = null;
                    }

                    //
                    // NOTE: Did we manage to find a suitable library path?
                    //
                    if (!String.IsNullOrEmpty(path))
                    {
                        SetLibraryPath(path, true);

                        return true;
                    }
                }
            }

            return false;
        }
        #endregion
    }
}
