/*
 * StringList.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Interfaces.Public;

namespace Eagle._Containers.Public
{
#if SERIALIZATION
    [Serializable()]
#endif
    [ObjectId("17df73b1-f419-498d-aeca-4be513e25310")]
    public sealed class StringList : List<string>, IStringList
    {
        #region Public Constants
        public static readonly string EmptyElement =
            Characters.OpenBrace.ToString() + Characters.CloseBrace.ToString();
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public StringList()
            : base()
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            int capacity
            )
            : base(capacity)
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            object value
            )
        {
            AddObjectOrObjects(value);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable collection
            )
        {
            AddObjects(collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable collection,
            bool @null
            )
        {
            AddObjects(collection, @null);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable<char> collection
            )
        {
            AddChars(collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable<string> collection
            )
            : base(collection)
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable<IPair<string>> collection
            )
        {
            Add(collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable<Argument> collection
            )
        {
            Add(collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IEnumerable<Result> collection
            )
        {
            Add(collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IList list,
            int startIndex
            )
        {
            Add(list, startIndex);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            IList<string> list,
            int startIndex
            )
        {
            Add(list, startIndex);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            params string[] strings
            )
            : base(strings)
        {
            // do nothing.
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            StringTransformCallback callback,
            IEnumerable<string> collection
            )
        {
            Add(callback, collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            StringTransformCallback callback,
            IEnumerable<Argument> collection
            )
        {
            Add(callback, collection);
        }

        ///////////////////////////////////////////////////////////////////////

        public StringList(
            StringTransformCallback callback,
            IEnumerable<Result> collection
            )
        {
            Add(callback, collection);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region ICloneable Members
        public object Clone()
        {
            return new StringList(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IStringList Members
        #region Add Methods
        public void Add(
            string key,
            string value
            )
        {
            base.Add(key);
            base.Add(value);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IList list,
            int startIndex
            )
        {
            for (int index = startIndex; index < list.Count; index++)
                base.Add(RuntimeOps.GetStringFromObject(list[index]));
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IList<string> list,
            int startIndex
            )
        {
            for (int index = startIndex; index < list.Count; index++)
                base.Add(list[index]);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IEnumerable<string> collection
            )
        {
            foreach (string item in collection)
                base.Add(item);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IDictionary<string, string> dictionary
            )
        {
            foreach (KeyValuePair<string, string> pair in dictionary)
                Add(pair.Key, pair.Value);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IEnumerable<Argument> collection
            )
        {
            foreach (Argument item in collection)
                base.Add(item);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IEnumerable<Result> collection
            )
        {
            foreach (Result item in collection)
                base.Add(item);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            IEnumerable<IPair<string>> collection
            )
        {
            foreach (IPair<string> item in collection)
                Add(item.X, item.Y);
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            StringTransformCallback callback,
            IEnumerable<string> collection
            )
        {
            foreach (string item in collection)
                base.Add(callback(item));
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            StringTransformCallback callback,
            IEnumerable<Argument> collection
            )
        {
            foreach (Argument item in collection)
                base.Add(callback(item));
        }

        ///////////////////////////////////////////////////////////////////////

        public void Add(
            StringTransformCallback callback,
            IEnumerable<Result> collection
            )
        {
            foreach (Result item in collection)
                base.Add(callback(item));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region ToString Methods
        public string ToString(
            bool empty
            )
        {
            if (empty)
            {
                return ToString();
            }
            else
            {
                StringList result = new StringList();

                foreach (string element in this)
                    if (!String.IsNullOrEmpty(element))
                        result.Add(element);

                return result.ToString();
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToString(
            string pattern,
            bool noCase
            )
        {
            return ToString(Characters.Space.ToString(), pattern, noCase);
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToString(
            string separator,
            string pattern,
            bool noCase
            )
        {
            return GenericOps<string>.ListToString(this, Index.Invalid,
                Index.Invalid, ToStringFlags.None, separator, pattern, noCase);
        }

        ///////////////////////////////////////////////////////////////////////

        public string ToRawString()
        {
            StringBuilder result = StringOps.NewStringBuilder();

            foreach (string element in this)
                result.Append(element);

            return result.ToString();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        public IStringList ToList()
        {
            return this; /* NOTE: Gee, that was easy. */
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Methods
        #region Add Methods
        public void Add(
            string item1,
            string item2,
            params string[] strings
            )
        {
            Add(item1, item2);
            Add(strings);
        }

        ///////////////////////////////////////////////////////////////////////

        public void AddChars(
            IEnumerable<char> collection
            )
        {
            foreach (char item in collection)
                base.Add(item.ToString());
        }

        ///////////////////////////////////////////////////////////////////////

        public void AddObjects(
            IEnumerable collection
            )
        {
            foreach (object item in collection)
                base.Add(RuntimeOps.GetStringFromObject(item));
        }

        ///////////////////////////////////////////////////////////////////////

        public void AddObjects(
            IEnumerable collection,
            bool @null
            )
        {
            foreach (object item in collection)
            {
                if (item != null)
                    base.Add(RuntimeOps.GetStringFromObject(item));
                else if (@null)
                    base.Add(null);
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public void AddObjectOrObjects(
            object value
            )
        {
            if (value != null)
            {
                IEnumerable enumerable = value as IEnumerable;

                if (enumerable != null)
                    AddObjects(enumerable, true);
                else
                    AddObjects(new object[] { value }, true);
            }
            else
            {
                base.Add(null);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Range Methods
        public void InsertRange(
            int index,
            IList list,
            int startIndex
            )
        {
            base.InsertRange(index, new StringList(list, startIndex));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Search Methods
        public bool Contains(
            string item,
            StringComparison comparisonType
            )
        {
            for (int index = 0; index < base.Count; index++)
                if (String.Equals(base[index], item, comparisonType))
                    return true;

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public int IndexOf(
            string value,
            int startIndex,
            StringComparison comparisonType
            )
        {
            for (int index = startIndex; index < base.Count; index++)
                if (String.Equals(base[index], value, comparisonType))
                    return index;

            return Index.Invalid;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Static Methods
        #region Range Methods
        public static StringList GetRange(
            IList list,
            int firstIndex
            )
        {
            return GetRange(list, firstIndex,
                (list != null) ? (list.Count - 1) : Index.Invalid);
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList GetRange(
            IList list,
            int firstIndex,
            int lastIndex
            )
        {
            StringList range = null;

            if (list != null)
            {
                range = new StringList();

                for (int index = firstIndex; index <= lastIndex; index++)
                    range.Add(RuntimeOps.GetStringFromObject(list[index]));
            }

            return range;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region ToString Methods
        public static string MakeList(
            params string[] strings
            )
        {
            return MakeList((IEnumerable<string>)strings);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string MakeList(
            IEnumerable<string> collection
            )
        {
            return new StringList(collection).ToString();
        }

        ///////////////////////////////////////////////////////////////////////

        public static string MakeList(
            params object[] objects
            )
        {
            return MakeList((IEnumerable)objects);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string MakeList(
            IEnumerable collection
            )
        {
            return new StringList(collection).ToString();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Factory Methods
        public static StringList FromString(
            string value
            )
        {
            Result error = null;

            return FromString(value, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList FromString(
            string value,
            ref Result error
            )
        {
            StringList list = null;

            //
            // TODO: *PERF* We cannot have this call to SplitList perform any
            //       caching because we do not know exactly what the resulting
            //       list will be used for.
            //
            if (Parser.SplitList(null, value, 0, Length.Invalid,
                    false, false, ref list, ref error) != ReturnCode.Ok)
            {
                list = null;
            }

            return list;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region System.Object Overrides
        public override string ToString()
        {
            return ToString(Characters.Space.ToString(), null, false);
        }
        #endregion
    }
}
