/*
 * GarudaInt.h -- Eagle Package for Tcl (Garuda)
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

#ifndef _GARUDA_INT_H_
#define _GARUDA_INT_H_

#define UNICODE_TEXT(x)				UNICODE_TEXT1(x)
#define UNICODE_TEXT1(x)			L##x

#define PACKAGE_UNICODE_NAME			UNICODE_TEXT(PACKAGE_NAME)
#define PACKAGE_UNICODE_VERSION			UNICODE_TEXT(PACKAGE_VERSION)
#define PACKAGE_PTR_FORMAT			UNICODE_TEXT("0x%p")
#define PACKAGE_RESULT_SIZE			(1024)
#define PACKAGE_CAN_LOG(a,b)			(((a) != NULL) && ((b) != NULL))

/*
 * NOTE: When the package is being compiled with the PACKAGE_DEBUG option
 *       enabled, we want to handle serious errors using either the Tcl_Panic
 *       function from the Tcl API or the printf function from the CRT.  Which
 *       of these functions gets used depends on the build configuration.  In
 *       the "Debug" build configuration, the Tcl_Panic function is used so
 *       that it can immediately abort the process.  In the "Release" build
 *       configuration, the printf function is used to report the error to the
 *       standard output channel, if available.  Currently, there are only two
 *       places where this macro is used and neither of them strictly require
 *       the process to be aborted; therefore, when the package is compiled
 *       without the PACKAGE_DEBUG option enabled, this macro does nothing.
 */

#ifdef PACKAGE_DEBUG
  #ifdef _DEBUG
    #define PACKAGE_PANIC(x)			Tcl_Panic x
  #else
    #define PACKAGE_PANIC(x)			printf x
  #endif
#else
  #define PACKAGE_PANIC(x)
#endif

/*
 * NOTE: These variable names are built using the base variable name strings
 *       defined in "Garuda.h".  The package name is prefixed to each variable
 *       name and the entire string is "converted" to Unicode.  These variables
 *       are always resolved relative to the global namespace.
 */

#define PACKAGE_UNICODE_ASSEMBLY_PATH_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(ASSEMBLY_PATH_VAR_NAME))

#define PACKAGE_UNICODE_TYPE_NAME_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(TYPE_NAME_VAR_NAME))

#define PACKAGE_UNICODE_STARTUP_METHOD_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(STARTUP_METHOD_VAR_NAME))

#define PACKAGE_UNICODE_CONTROL_METHOD_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(CONTROL_METHOD_VAR_NAME))

#define PACKAGE_UNICODE_DETACH_METHOD_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(DETACH_METHOD_VAR_NAME))

#define PACKAGE_UNICODE_SHUTDOWN_METHOD_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(SHUTDOWN_METHOD_VAR_NAME))

#define PACKAGE_UNICODE_METHOD_ARGUMENTS_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(METHOD_ARGUMENTS_VAR_NAME))

#define PACKAGE_UNICODE_METHOD_FLAGS_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(METHOD_FLAGS_VAR_NAME))

#define PACKAGE_UNICODE_VERBOSE_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(VERBOSE_VAR_NAME))

#define PACKAGE_UNICODE_START_CLR_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(START_CLR_VAR_NAME))

#define PACKAGE_UNICODE_START_BRIDGE_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(START_BRIDGE_VAR_NAME))

#define PACKAGE_UNICODE_LOG_COMMAND_VAR_NAME \
    JOIN(PACKAGE_UNICODE_NAME, UNICODE_TEXT(LOG_COMMAND_VAR_NAME))

/*
 * NOTE: This is the latest version of the CLR that we know about.  This is
 *       the value that will be passed to the GetRuntime method of the
 *       ICLRMetaHost interface (which will only be used when the CLR_40
 *       compile-time option is enabled).  For now, only configure use of the
 *       latest version of the CLR if we are compiling with the MSVC compiler
 *       that shipped with Visual Studio 2010 or higher.
 */

#if defined(USE_CLR_40)
  #define CLR_MODULE_NAME			"MSCorEE"
  #define CLR_PROC_NAME				"CLRCreateInstance"
  #define CLR_VERSION_LATEST			"v4.0.30319"
#endif

/*
 * Flag values for the <pkg>_Unload callback function (Tcl 8.5+).
 */

#if !defined(TCL_UNLOAD_DETACH_FROM_INTERPRETER)
  #define TCL_UNLOAD_DETACH_FROM_INTERPRETER	(1<<0)
#endif

#if !defined(TCL_UNLOAD_DETACH_FROM_PROCESS)
  #define TCL_UNLOAD_DETACH_FROM_PROCESS	(1<<1)
#endif

/*
 * HACK: This flag means that the <pkg>_Unload callback function is being
 *       called from the <pkg>_Init function to cleanup due to a package load
 *       failure.  This flag is NOT defined by Tcl and MUST NOT conflict with
 *       the unloading related flags defined by Tcl in "tcl.h".
 */

#if !defined(TCL_UNLOAD_FROM_INIT)
  #define TCL_UNLOAD_FROM_INIT			(1<<2)
#endif

/*
 * NOTE: This enumeration contains the "types" of [and flags for] the CLR
 *       methods this package knows how to deal with.
 */

typedef enum {
    METHOD_NONE = 0x0,		    /* Unknown method type with no flags. */
    METHOD_TYPE_DEMAND = 0x1,	    /* Used for any method executed on-demand
				     * via the Tcl command interface. */
    METHOD_TYPE_STARTUP = 0x2,	    /* Startup the bridge between Eagle and
				     * Tcl. */
    METHOD_TYPE_CONTROL = 0x4,	    /* Issue a control directive to the bridge
				     * between Eagle and Tcl. */
    METHOD_TYPE_DETACH = 0x8,	    /* Detach a Tcl interpreter from the bridge
				     * between Eagle and Tcl (i.e. make the
				     * bridge "forget" about it because it was
				     * deleted, the package was unloaded from
				     * it, etc). */
    METHOD_TYPE_SHUTDOWN = 0x10,    /* Shutdown the bridge between Eagle and
				     * Tcl. */
    METHOD_TYPE_MASK = 0x1F,	    /* Used for limiting the method flags to
				     * [one or more] of the type values. */
    METHOD_PROTOCOL_V1 = 0x20,	    /* If set, the protocol version indicator,
				     * the Tcl library module handle, the Tcl
				     * interpreter pointer, and Tcl interpreter
				     * safety indicator will be converted to
				     * strings and prepended to the configured
				     * and specified method arguments, if any. */
    METHOD_LOG_EXECUTE = 0x40,	    /* If set, method execution will be logged
				     * using the configured Tcl command name. */
    METHOD_STRICT_CLR = 0x80,	    /* If set, method execution will be
				     * considered a failure if the CLR has not
				     * been started by this package. */
    METHOD_STRICT_RETURN = 0x100,   /* If set, the return value from the CLR
				     * method must be TCL_OK to be considered
				     * successful. */

    /*
     * NOTE: These are the "standard" flag combinations used by this package to
     *       execute CLR methods at specific stages in the package lifecycle.
     */

    METHOD_VIA_DEMAND = METHOD_LOG_EXECUTE | METHOD_STRICT_CLR,

    METHOD_VIA_LOAD = METHOD_PROTOCOL_V1 | METHOD_LOG_EXECUTE |
		      METHOD_STRICT_CLR | METHOD_STRICT_RETURN,

    METHOD_VIA_COMMAND = METHOD_PROTOCOL_V1 | METHOD_LOG_EXECUTE |
			 METHOD_STRICT_CLR | METHOD_STRICT_RETURN,

    METHOD_VIA_UNLOAD = METHOD_PROTOCOL_V1 | METHOD_LOG_EXECUTE |
			METHOD_STRICT_RETURN,
} MethodFlags;

/*
 * NOTE: This structure contains the information used by this package to execute
 *       a CLR method.
 */

typedef struct ClrMethodInfo {
    size_t sizeOf;	    /* The size of this structure, in bytes. */
    LPCWSTR assemblyPath;   /* The fully qualified path and file name of the
			     * assembly containing the method to execute. */
    LPCWSTR typeName;	    /* The fully qualified type name of the type
			     * containing the method to execute. */
    LPCWSTR methodName;	    /* The method name to execute. */
    LPCWSTR argument;	    /* The argument string to be passed to the method.
			     * This SHOULD be a well-formed Tcl list. */
} ClrMethodInfo;

/*
 * NOTE: This structure contains the various configuration settings used by this
 *       package.  Currently, all of these settings are read from Tcl variables;
 *       however, this may not always be the case.
 */

typedef struct ClrConfigInfo {
    size_t sizeOf;		    /* The size of this structure, in bytes. */
    ClrMethodInfo *pStartupMethod;  /* The [cached] CLR method used to startup
				     * the bridge between Eagle and Tcl, if any. */
    ClrMethodInfo *pControlMethod;  /* The [cached] CLR method used to issue
				     * control directives to the bridge between
				     * Eagle and Tcl, if any. */
    ClrMethodInfo *pDetachMethod;   /* The [cached] CLR method used to detach a
				     * Tcl interpreter from the bridge between
				     * Eagle and Tcl, if any. */
    ClrMethodInfo *pShutdownMethod; /* The [cached] CLR method used to shutdown
				     * the bridge between Eagle and Tcl, if any. */
    LPCWSTR logCommand;		    /* The name of the Tcl command to use for
				     * logging purposes (e.g. tclLog). */
    MethodFlags methodFlags;	    /* The method flags to use when executing
				     * any of the CLR methods. */
    BOOL bStartClr;		    /* Start the CLR immediately upon loading
				     * the package? */
    BOOL bStartBridge;		    /* Start the bridge between Eagle and Tcl
				     * immediately upon loading the package? */
} ClrConfigInfo;

#endif /* _GARUDA_INT_H_ */
